#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.cli.checks`."""
import pytest

import orion.core
from orion.core.cli.checks.creation import CreationStage
from orion.core.cli.checks.operations import OperationsStage
from orion.core.cli.checks.presence import PresenceStage
from orion.core.io.database.mongodb import MongoDB
import orion.core.io.experiment_builder as experiment_builder
import orion.core.utils.backward as backward
from orion.core.utils.exceptions import CheckError


@pytest.fixture
def config():
    """Return a basic database configuration."""
    return {'database': {'host': 'localhost', 'type': 'mongodb', 'name': 'user'}}


@pytest.fixture
def presence():
    """Return a PresenceStage instance."""
    return PresenceStage([])


@pytest.fixture
def creation(create_db_instance):
    """Return a CreationStage instance."""
    stage = CreationStage(None)
    stage.instance = create_db_instance
    return stage


@pytest.fixture
def operation(creation):
    """Return an OperationStage instance."""
    return OperationsStage(creation)


@pytest.fixture
def clean_test(database):
    """Remove the test collection from database."""
    database.test.drop()


def test_check_default_config_pass(monkeypatch, presence, config):
    """Check if the default config test works."""
    def mock_default_config(self):
        return config

    monkeypatch.setattr(orion.core.config.__class__, 'to_dict', mock_default_config)

    result, msg = presence.check_default_config()

    assert result == "Success"
    assert msg == ""
    assert presence.db_config == config['storage']['database']


def test_check_default_config_skip(monkeypatch, presence):
    """Check if test returns skip if no default config is found."""
    def mock_default_config(self):
        return {}

    monkeypatch.setattr(orion.core.config.__class__, 'to_dict', mock_default_config)

    result, msg = presence.check_default_config()
    assert result == "Skipping"
    assert "No" in msg
    assert presence.db_config == {}


def test_config_file_config_pass(monkeypatch, presence, config):
    """Check if test passes with valid configuration."""
    def mock_file_config(cmdargs):
        backward.update_db_config(config)
        return config

    monkeypatch.setattr(experiment_builder, "get_cmd_config", mock_file_config)

    result, msg = presence.check_configuration_file()

    assert result == "Success"
    assert msg == ""
    assert presence.db_config == config['storage']['database']


def test_config_file_fails_missing_config(monkeypatch, presence, config):
    """Check if test fails with missing configuration."""
    def mock_file_config(cmdargs):
        return {}

    monkeypatch.setattr(experiment_builder, "get_cmd_config", mock_file_config)

    status, msg = presence.check_configuration_file()

    assert status == "Skipping"
    assert "Missing" in msg
    assert presence.db_config == {}


def test_config_file_fails_missing_database(monkeypatch, presence, config):
    """Check if test fails with missing database configuration."""
    def mock_file_config(cmdargs):
        return {'algorithm': 'asha'}

    monkeypatch.setattr(experiment_builder, "get_cmd_config", mock_file_config)

    status, msg = presence.check_configuration_file()

    assert status == "Skipping"
    assert "No database" in msg
    assert presence.db_config == {}


def test_config_file_fails_missing_value(monkeypatch, presence, config):
    """Check if test fails with missing value in database configuration."""
    def mock_file_config(cmdargs):
        return {'storage': {'database': {}}}

    monkeypatch.setattr(experiment_builder, "get_cmd_config", mock_file_config)

    status, msg = presence.check_configuration_file()

    assert status == "Skipping"
    assert "No configuration" in msg
    assert presence.db_config == {}


def test_config_file_skips(monkeypatch, presence, config):
    """Check if test skips when another configuration is present."""
    def mock_file_config(self):
        return {}

    presence.db_config = config['database']
    monkeypatch.setattr(experiment_builder, "get_cmd_config", mock_file_config)

    result, msg = presence.check_configuration_file()

    assert result == "Skipping"
    assert presence.db_config == config['database']


@pytest.mark.usefixtures('null_db_instances')
def test_creation_pass(presence, config):
    """Check if test passes with valid database configuration."""
    presence.db_config = config['database']
    creation = CreationStage(presence)

    result, msg = creation.check_database_creation()

    assert result == "Success"
    assert msg == ""
    assert creation.instance is not None


@pytest.mark.usefixtures('null_db_instances')
def test_creation_fails(monkeypatch, presence, config):
    """Check if test fails when not connected."""
    presence.db_config = config['database']
    creation = CreationStage(presence)

    monkeypatch.setattr(MongoDB, "is_connected", False)

    with pytest.raises(CheckError) as ex:
        creation.check_database_creation()

    assert "failed" in str(ex.value)


def test_operation_write_pass(operation):
    """Check if test passes when write operation works."""
    result, msg = operation.check_write()

    assert result == "Success"
    assert msg == ""


def test_operation_write_fails(monkeypatch, operation):
    """Check if test fails when write operation fails."""
    def mock_write(one, two):
        raise RuntimeError("Not working")

    monkeypatch.setattr(operation.c_stage.instance, "write", mock_write)

    with pytest.raises(CheckError) as ex:
        operation.check_write()

    assert "Not working" in str(ex.value)


def test_operation_read_pass(operation, clean_test):
    """Check if test passes when read operation works."""
    operation.c_stage.instance.write('test', {'index': 'value'})
    result, msg = operation.check_read()

    assert result == "Success"
    assert msg == ""


def test_operation_read_fail_not_working(monkeypatch, operation):
    """Check if test fails when read operation fails."""
    def mock_read(one, two):
        raise RuntimeError("Not working")

    monkeypatch.setattr(operation.c_stage.instance, "read", mock_read)

    with pytest.raises(CheckError) as ex:
        operation.check_read()

    assert "Not working" in str(ex.value)


def test_operation_read_fail_unexpected_value(operation, clean_test):
    """Check if test fails on unexpected read value."""
    operation.c_stage.instance.write('test', {'index': 'value2'})

    with pytest.raises(CheckError) as ex:
        operation.check_read()

    assert "value" in str(ex.value)


def test_operation_count_pass(operation, clean_test):
    """Check if test passes when count operation works."""
    operation.c_stage.instance.write('test', {'index': 'value'})
    result, msg = operation.check_count()

    assert result == "Success"
    assert msg == ""


def test_operation_count_fails(monkeypatch, operation, clean_test):
    """Check if test fails when count operation fails."""
    operation.c_stage.instance.write('test', {'index': 'value'})
    operation.c_stage.instance.write('test', {'index': 'value'})
    with pytest.raises(CheckError) as ex:
        operation.check_count()

    assert "2" in str(ex.value)


def test_operation_remove_pass(operation, clean_test):
    """Check if test passes when remove operation works."""
    operation.c_stage.instance.write('test', {'index': 'value'})
    result, msg = operation.check_remove()

    assert result == "Success"
    assert msg == ""


def test_operation_remove_fails(monkeypatch, operation, clean_test, database):
    """Check if test fails when remove operation fails."""
    operation.c_stage.instance.write('test', {'index': 'value'})
    operation.c_stage.instance.write('test', {'index': 'value'})

    def mock_remove(one, two):
        database.test.delete_one({'index': 'value'})

    monkeypatch.setattr(operation.c_stage.instance, "remove", mock_remove)

    with pytest.raises(CheckError) as ex:
        operation.check_remove()

    assert "1" in str(ex.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Simple one dimensional example for a possible user's script."""
import argparse


def execute():
    """Execute a simple pipeline as an example."""
    # 1. Receive inputs as you want
    parser = argparse.ArgumentParser()
    parser.add_argument('-x', type=float, required=True)

    raise RuntimeError


if __name__ == "__main__":
    execute()



"""Collection of tests for :mod:`orion.plotting.backend_plotly`."""
import plotly
import pytest

from orion.core.worker.experiment import ExperimentView
from orion.plotting.base import regret
from orion.testing import create_experiment

config = dict(
    name='experiment-name',
    space={'x': 'uniform(0, 200)'},
    metadata={'user': 'test-user',
              'orion_version': 'XYZ',
              'VCS': {"type": "git",
                      "is_dirty": False,
                      "HEAD_sha": "test",
                      "active_branch": None,
                      "diff_sha": "diff"}},
    version=1,
    pool_size=1,
    max_trials=10,
    working_dir='',
    algorithms={'random': {'seed': 1}},
    producer={'strategy': 'NoParallelStrategy'},
)

trial_config = {
    'experiment': 0,
    'status': 'completed',
    'worker': None,
    'start_time': None,
    'end_time': None,
    'heartbeat': None,
    'results': [],
    'params': []
}


def assert_regret_plot(plot):
    """Checks the layout of a regret plot"""
    assert plot.layout.title.text == "Regret for experiment 'experiment-name'"
    assert plot.layout.xaxis.title.text == "Trials ordered by suggested time"
    assert plot.layout.yaxis.title.text == "Objective 'loss'"

    trace1 = plot.data[0]
    assert trace1.type == "scatter"
    assert trace1.name == "trials"
    assert trace1.mode == "markers"
    assert len(trace1.y) == 1
    assert not trace1.x

    trace2 = plot.data[1]
    assert trace2.type == "scatter"
    assert trace2.name == "best-to-date"
    assert trace2.mode == "lines"
    assert len(trace2.y) == 1
    assert not trace2.x


class TestRegret:
    """Tests the ``regret()`` method provided by the plotly backend"""

    def test_requires_argument(self):
        """Tests that the experiment data are required."""
        with pytest.raises(ValueError):
            regret(None)

    def test_returns_plotly_object(self):
        """Tests that the plotly backend returns a plotly object"""
        with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
            plot = regret(experiment)

        assert type(plot) is plotly.graph_objects.Figure

    def test_graph_layout(self):
        """Tests the layout of the plot"""
        with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
            plot = regret(experiment)

        assert_regret_plot(plot)

    def test_experiment_worker_as_parameter(self):
        """Tests that ``Experiment`` is a valide parameter"""
        with create_experiment(config, trial_config, ['completed']) as (_, experiment, _):
            plot = regret(experiment)

        assert_regret_plot(plot)

    def test_experiment_view_as_parameter(self):
        """Tests that ``ExperimentView`` is a valid parameter"""
        with create_experiment(config, trial_config, ['completed']) as (_, experiment, _):
            plot = regret(ExperimentView(experiment))

        assert_regret_plot(plot)

    def test_ignore_uncompleted_statuses(self):
        """Tests that uncompleted statuses are ignored"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            plot = regret(experiment)

        assert_regret_plot(plot)

    def test_unsupported_order_key(self):
        """Tests that unsupported order keys are rejected"""
        with create_experiment(config, trial_config) as (_, _, experiment):
            with pytest.raises(ValueError):
                regret(experiment, order_by='unsupported')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.evc.conflicts`."""

import pprint

import pytest

from orion.algo.space import Dimension
from orion.core import evc
from orion.core.evc import conflicts as conflict
import orion.core.utils.backward as backward


@pytest.fixture
def conflicts(new_dimension_conflict, changed_dimension_conflict, missing_dimension_conflict,
              algorithm_conflict, code_conflict, experiment_name_conflict):
    """Create a container for conflicts with one of each types for testing purposes"""
    conflicts = evc.conflicts.Conflicts()
    conflicts.register(new_dimension_conflict)
    conflicts.register(changed_dimension_conflict)
    conflicts.register(missing_dimension_conflict)
    conflicts.register(algorithm_conflict)
    conflicts.register(code_conflict)
    conflicts.register(experiment_name_conflict)
    return conflicts


class TestNewDimensionConflict(object):
    """Tests methods related to new dimension conflicts"""

    def test_try_resolve_no_default(self, new_dimension_conflict):
        """Verify that resolution is achievable without default value"""
        assert not new_dimension_conflict.is_resolved
        resolution = new_dimension_conflict.try_resolve()
        assert new_dimension_conflict.is_resolved
        assert isinstance(resolution, new_dimension_conflict.AddDimensionResolution)
        assert resolution.default_value is Dimension.NO_DEFAULT_VALUE

    def test_try_resolve_default_from_dim(self, new_dimension_with_default_conflict):
        """Verify that resolution includes dimension's default value if none is provided"""
        assert not new_dimension_with_default_conflict.is_resolved
        resolution = new_dimension_with_default_conflict.try_resolve()
        assert new_dimension_with_default_conflict.is_resolved
        assert isinstance(resolution, new_dimension_with_default_conflict.AddDimensionResolution)
        assert (resolution.default_value ==
                new_dimension_with_default_conflict.dimension.default_value)

    def test_try_resolve_default(self, new_dimension_conflict):
        """Verify that resolution includes default value if provided"""
        default_value = 1.2
        assert not new_dimension_conflict.is_resolved
        resolution = new_dimension_conflict.try_resolve(default_value)
        assert new_dimension_conflict.is_resolved
        assert isinstance(resolution, new_dimension_conflict.AddDimensionResolution)
        assert (resolution.default_value == default_value)

    def test_try_resolve_default_over_dim(self, new_dimension_with_default_conflict):
        """Verify that resolution overwrites dimension's default value if one is provided"""
        default_value = 1.2
        assert not new_dimension_with_default_conflict.is_resolved
        resolution = new_dimension_with_default_conflict.try_resolve(default_value)
        assert new_dimension_with_default_conflict.is_resolved
        assert isinstance(resolution, new_dimension_with_default_conflict.AddDimensionResolution)
        assert (resolution.default_value == default_value)

    def test_try_resolve_bad_default(self, new_dimension_conflict):
        """Verify that resolution fails if default value is invalid"""
        assert not new_dimension_conflict.is_resolved
        with pytest.raises(ValueError) as exc:
            new_dimension_conflict.try_resolve('bad-default')
        assert "could not convert string to float: 'bad-default'" in str(exc.value)

    def test_try_resolve_twice(self, new_dimension_conflict):
        """Verify that conflict cannot be resolved twice"""
        assert not new_dimension_conflict.is_resolved
        assert isinstance(new_dimension_conflict.try_resolve(),
                          new_dimension_conflict.AddDimensionResolution)
        assert new_dimension_conflict.is_resolved
        assert new_dimension_conflict.try_resolve() is None

    def test_repr(self, new_dimension_conflict):
        """Verify the representation of conflict for user interface"""
        assert repr(new_dimension_conflict) == "New new"


class TestChangedDimensionConflict(object):
    """Tests methods related to changed dimension prior conflicts"""

    def test_try_resolve_twice(self, changed_dimension_conflict):
        """Verify that conflict cannot be resolved twice"""
        assert not changed_dimension_conflict.is_resolved
        assert isinstance(changed_dimension_conflict.try_resolve(),
                          changed_dimension_conflict.ChangeDimensionResolution)
        assert changed_dimension_conflict.is_resolved
        assert changed_dimension_conflict.try_resolve() is None

    def test_try_resolve(self, changed_dimension_conflict):
        """Verify that resolution is achievable without any input"""
        assert not changed_dimension_conflict.is_resolved
        resolution = changed_dimension_conflict.try_resolve()
        assert isinstance(resolution, changed_dimension_conflict.ChangeDimensionResolution)
        assert changed_dimension_conflict.is_resolved
        assert resolution.conflict is changed_dimension_conflict

    def test_repr(self, changed_dimension_conflict):
        """Verify the representation of conflict for user interface"""
        assert (repr(changed_dimension_conflict) ==
                "changed~uniform(-10, 10) != changed~normal(0, 2)")


class TestMissingDimensionConflict(object):
    """Tests methods related to missing dimension conflicts"""

    def test_try_resolve_no_default(self, missing_dimension_conflict):
        """Verify that resolution is achievable without default value provided"""
        assert not missing_dimension_conflict.is_resolved
        resolution = missing_dimension_conflict.try_resolve()
        assert missing_dimension_conflict.is_resolved
        assert isinstance(resolution, missing_dimension_conflict.RemoveDimensionResolution)
        assert resolution.default_value is Dimension.NO_DEFAULT_VALUE

    def test_try_resolve_default_from_dim(self, missing_dimension_with_default_conflict):
        """Verify that resolution uses default value from dim when no provided by user"""
        assert not missing_dimension_with_default_conflict.is_resolved
        resolution = missing_dimension_with_default_conflict.try_resolve()
        assert missing_dimension_with_default_conflict.is_resolved
        assert isinstance(resolution,
                          missing_dimension_with_default_conflict.RemoveDimensionResolution)
        assert (resolution.default_value ==
                missing_dimension_with_default_conflict.dimension.default_value)

    def test_try_resolve_default(self, missing_dimension_conflict):
        """Verify that resolution uses default value provided by user"""
        default_value = 1.2
        assert not missing_dimension_conflict.is_resolved
        resolution = missing_dimension_conflict.try_resolve(default_value=default_value)
        assert missing_dimension_conflict.is_resolved
        assert isinstance(resolution, missing_dimension_conflict.RemoveDimensionResolution)
        assert (resolution.default_value == default_value)

    def test_try_resolve_default_over_dim(self, missing_dimension_with_default_conflict):
        """Verify that resolution overwrite dimension's default value when user provide one"""
        default_value = 1.2
        assert not missing_dimension_with_default_conflict.is_resolved
        resolution = missing_dimension_with_default_conflict.try_resolve(
            default_value=default_value)
        assert missing_dimension_with_default_conflict.is_resolved
        assert isinstance(resolution,
                          missing_dimension_with_default_conflict.RemoveDimensionResolution)
        assert (resolution.default_value == default_value)

    def test_try_resolve_bad_default(self, missing_dimension_conflict):
        """Verify that resolution fails if default value is invalid"""
        assert not missing_dimension_conflict.is_resolved
        with pytest.raises(ValueError) as exc:
            missing_dimension_conflict.try_resolve(default_value='-100')
        assert "Default value `-100.0` is outside" in str(exc.value)

    def test_try_resolve_renaming(self, missing_dimension_conflict, new_dimension_conflict):
        """Verify that resolution is a renaming when new_dimension_conflict is provided"""
        assert not missing_dimension_conflict.is_resolved
        resolution = missing_dimension_conflict.try_resolve(new_dimension_conflict)
        assert isinstance(resolution, missing_dimension_conflict.RenameDimensionResolution)
        assert resolution.conflict is missing_dimension_conflict
        assert resolution.new_dimension_conflict is new_dimension_conflict

    def test_try_resolve_twice(self, missing_dimension_conflict):
        """Verify that conflict cannot be resolved twice"""
        assert not missing_dimension_conflict.is_resolved
        assert isinstance(missing_dimension_conflict.try_resolve(),
                          missing_dimension_conflict.RemoveDimensionResolution)
        assert missing_dimension_conflict.is_resolved
        assert missing_dimension_conflict.try_resolve() is None

    def test_repr(self, missing_dimension_conflict):
        """Verify the representation of conflict for user interface"""
        assert repr(missing_dimension_conflict) == "Missing missing"


class TestAlgorithmConflict(object):
    """Tests methods related to algorithm configuration conflicts"""

    def test_try_resolve_twice(self, algorithm_conflict):
        """Verify that conflict cannot be resolved twice"""
        assert not algorithm_conflict.is_resolved
        assert isinstance(algorithm_conflict.try_resolve(),
                          algorithm_conflict.AlgorithmResolution)
        assert algorithm_conflict.is_resolved
        assert algorithm_conflict.try_resolve() is None

    def test_try_resolve(self, algorithm_conflict):
        """Verify that resolution is achievable without any input"""
        assert not algorithm_conflict.is_resolved
        resolution = algorithm_conflict.try_resolve()
        assert isinstance(resolution, algorithm_conflict.AlgorithmResolution)
        assert algorithm_conflict.is_resolved
        assert resolution.conflict is algorithm_conflict

    def test_repr(self, old_config, new_config, algorithm_conflict):
        """Verify the representation of conflict for user interface"""
        repr_baseline = "{0}\n   !=\n{1}".format(pprint.pformat(old_config['algorithms']),
                                                 pprint.pformat(new_config['algorithms']))
        assert repr(algorithm_conflict) == repr_baseline


class TestCodeConflict(object):
    """Tests methods related to code conflicts"""

    def test_try_resolve_twice(self, code_conflict):
        """Verify that conflict cannot be resolved twice"""
        assert not code_conflict.is_resolved
        assert isinstance(code_conflict.try_resolve(evc.adapters.CodeChange.BREAK),
                          code_conflict.CodeResolution)
        assert code_conflict.is_resolved
        assert code_conflict.try_resolve() is None

    def test_try_resolve(self, code_conflict):
        """Verify that resolution is achievable with valid code change-type input"""
        assert not code_conflict.is_resolved
        resolution = code_conflict.try_resolve(evc.adapters.CodeChange.UNSURE)
        assert isinstance(resolution, code_conflict.CodeResolution)
        assert code_conflict.is_resolved
        assert resolution.conflict is code_conflict
        assert resolution.type == evc.adapters.CodeChange.UNSURE

    def test_try_resolve_bad_code(self, code_conflict):
        """Verify that resolution fails if code is invalid"""
        assert not code_conflict.is_resolved
        with pytest.raises(ValueError) as exc:
            code_conflict.try_resolve()
        assert "Invalid code change type 'None'" in str(exc.value)

        with pytest.raises(ValueError) as exc:
            code_conflict.try_resolve("bad-change-type")
        assert "Invalid code change type 'bad-change-type'" in str(exc.value)

    @pytest.mark.usefixtures("mock_infer_versioning_metadata")
    def test_repr(self, code_conflict):
        """Verify the representation of conflict for user interface"""
        assert repr(code_conflict) == "Old hash commit '{'HEAD_sha': 'test', "\
                                      "'active_branch': None, 'diff_sha': 'diff', "\
                                      "'is_dirty': False, 'type': 'git'}'  "\
                                      "!= new hash commit ''to be changed''"

    def test_hash_commit_compar(self):
        """Test that old config hash commit evals to empty."""
        old_config = {'metadata': {'VCS': {}}}
        new_config = {'metadata': {'VCS': {}}}

        assert list(conflict.CodeConflict.detect(old_config, new_config)) == []


class TestCommandLineConflict(object):
    """Tests methods related to code conflicts"""

    def test_try_resolve_twice(self, cli_conflict):
        """Verify that conflict cannot be resolved twice"""
        assert not cli_conflict.is_resolved
        assert isinstance(cli_conflict.try_resolve(evc.adapters.CommandLineChange.BREAK),
                          cli_conflict.CommandLineResolution)
        assert cli_conflict.is_resolved
        assert cli_conflict.try_resolve() is None

    def test_try_resolve(self, cli_conflict):
        """Verify that resolution is achievable with valid cli change-type input"""
        assert not cli_conflict.is_resolved
        resolution = cli_conflict.try_resolve(evc.adapters.CommandLineChange.UNSURE)
        assert isinstance(resolution, cli_conflict.CommandLineResolution)
        assert cli_conflict.is_resolved
        assert resolution.conflict is cli_conflict
        assert resolution.type == evc.adapters.CommandLineChange.UNSURE

    def test_try_resolve_bad_cli(self, cli_conflict):
        """Verify that resolution fails if cli is invalid"""
        assert not cli_conflict.is_resolved
        with pytest.raises(ValueError) as exc:
            cli_conflict.try_resolve()
        assert "Invalid cli change type 'None'" in str(exc.value)

        with pytest.raises(ValueError) as exc:
            cli_conflict.try_resolve("bad-change-type")
        assert "Invalid cli change type 'bad-change-type'" in str(exc.value)

    def test_repr(self, cli_conflict):
        """Verify the representation of conflict for user interface"""
        assert (
            repr(cli_conflict) == (
                "Old arguments '_pos_0 abs_path/black_box.py' != "
                "new arguments '_pos_0 abs_path/black_box.py bool-arg True some-new args'"))


class TestScriptConfigConflict(object):
    """Tests methods related to code conflicts"""

    def test_try_resolve_twice(self, config_conflict):
        """Verify that conflict cannot be resolved twice"""
        assert not config_conflict.is_resolved
        assert isinstance(config_conflict.try_resolve(evc.adapters.ScriptConfigChange.BREAK),
                          config_conflict.ScriptConfigResolution)
        assert config_conflict.is_resolved
        assert config_conflict.try_resolve() is None

    def test_try_resolve(self, config_conflict):
        """Verify that resolution is achievable with valid config change-type input"""
        assert not config_conflict.is_resolved
        resolution = config_conflict.try_resolve(evc.adapters.ScriptConfigChange.UNSURE)
        assert isinstance(resolution, config_conflict.ScriptConfigResolution)
        assert config_conflict.is_resolved
        assert resolution.conflict is config_conflict
        assert resolution.type == evc.adapters.ScriptConfigChange.UNSURE

    def test_try_resolve_bad_config(self, config_conflict):
        """Verify that resolution fails if config is invalid"""
        assert not config_conflict.is_resolved
        with pytest.raises(ValueError) as exc:
            config_conflict.try_resolve()
        assert "Invalid script's config change type 'None'" in str(exc.value)

        with pytest.raises(ValueError) as exc:
            config_conflict.try_resolve("bad-change-type")
        assert "Invalid script's config change type 'bad-change-type'" in str(exc.value)

    def test_repr(self, config_conflict):
        """Verify the representation of conflict for user interface"""
        assert repr(config_conflict) == "Script's configuration file changed"

    def test_comparison(self, yaml_config, yaml_diff_config):
        """Test that different configs are detected as conflict."""
        old_config = {'metadata': {'user_args': yaml_config}}
        new_config = {'metadata': {'user_args': yaml_diff_config}}

        backward.populate_space(old_config)
        backward.populate_space(new_config)

        conflicts = list(conflict.ScriptConfigConflict.detect(old_config, new_config))
        assert len(conflicts) == 1

    def test_comparison_idem(self, yaml_config):
        """Test that identical configs are not detected as conflict."""
        old_config = {'metadata': {'user_args': yaml_config}}
        new_config = {'metadata': {'user_args': yaml_config + ['--other', 'args']}}

        backward.populate_space(old_config)
        backward.populate_space(new_config)

        assert list(conflict.ScriptConfigConflict.detect(old_config, new_config)) == []


@pytest.mark.usefixtures("create_db_instance")
class TestExperimentNameConflict(object):
    """Tests methods related to experiment name conflicts"""

    def test_try_resolve_twice(self, experiment_name_conflict):
        """Verify that conflict cannot be resolved twice"""
        assert not experiment_name_conflict.is_resolved
        assert isinstance(experiment_name_conflict.try_resolve("dummy"),
                          experiment_name_conflict.ExperimentNameResolution)
        assert experiment_name_conflict.is_resolved
        assert experiment_name_conflict.try_resolve() is None

    def test_try_resolve(self, experiment_name_conflict):
        """Verify that resolution is achievable with a valid name"""
        new_name = "dummy"
        assert not experiment_name_conflict.is_resolved
        resolution = experiment_name_conflict.try_resolve(new_name)
        assert isinstance(resolution, experiment_name_conflict.ExperimentNameResolution)
        assert experiment_name_conflict.is_resolved
        assert resolution.conflict is experiment_name_conflict
        assert resolution.new_name == new_name

    def test_branch_w_existing_exp(self, existing_exp_conflict):
        """Test branching when an existing experiment with the new name already exists"""
        with pytest.raises(ValueError) as exc:
            existing_exp_conflict.try_resolve("dummy")

        assert "Cannot" in str(exc.value)

    def test_conflict_exp_no_child(self, exp_no_child_conflict):
        """Verify the version number is incremented when exp has no child."""
        new_name = "test"
        assert not exp_no_child_conflict.is_resolved
        resolution = exp_no_child_conflict.try_resolve(new_name)
        assert isinstance(resolution, exp_no_child_conflict.ExperimentNameResolution)
        assert exp_no_child_conflict.is_resolved
        assert resolution.conflict is exp_no_child_conflict
        assert resolution.old_version == 1
        assert resolution.new_version == 2

    def test_conflict_exp_w_child(self, exp_w_child_conflict):
        """Verify the version number is incremented from child when exp has a child."""
        new_name = "test"
        assert not exp_w_child_conflict.is_resolved
        resolution = exp_w_child_conflict.try_resolve(new_name)
        assert isinstance(resolution, exp_w_child_conflict.ExperimentNameResolution)
        assert exp_w_child_conflict.is_resolved
        assert resolution.conflict is exp_w_child_conflict
        assert resolution.new_version == 3

    def test_conflict_exp_w_child_as_parent(self, exp_w_child_as_parent_conflict):
        """Verify that an error is raised when trying to branch from parent."""
        new_name = "test"
        with pytest.raises(ValueError) as exc:
            exp_w_child_as_parent_conflict.try_resolve(new_name)

        assert "Experiment name" in str(exc.value)

    def test_conflict_exp_renamed(self, exp_w_child_conflict):
        """Verify the version number is not incremented when exp is renamed."""
        # It increments from child
        new_name = "test2"
        assert not exp_w_child_conflict.is_resolved
        resolution = exp_w_child_conflict.try_resolve(new_name)
        assert isinstance(resolution, exp_w_child_conflict.ExperimentNameResolution)
        assert exp_w_child_conflict.is_resolved
        assert resolution.conflict is exp_w_child_conflict
        assert resolution.old_version == 2
        assert resolution.new_version == 1

    def test_repr(self, experiment_name_conflict):
        """Verify the representation of conflict for user interface"""
        assert (repr(experiment_name_conflict) ==
                "Experiment name 'test' already exist with version '1'")


class TestConflicts(object):
    """Tests for Conflicts container, for getter, deprecation and resolve methods"""

    def test_register(self, conflicts):
        """Verify that conflicts are properly registerd"""
        assert len(conflicts.conflicts) == 6
        isinstance(conflicts.conflicts[0], evc.conflicts.NewDimensionConflict)
        isinstance(conflicts.conflicts[-1], evc.conflicts.ExperimentNameConflict)

    def test_get(self, conflicts):
        """Verify that bare get() fetches all conflicts"""
        assert len(conflicts.get()) == 6
        isinstance(conflicts.get()[0], evc.conflicts.NewDimensionConflict)
        isinstance(conflicts.get()[-1], evc.conflicts.ExperimentNameConflict)

    def test_get_one_type(self, conflicts):
        """Verify that get() fetches proper given type"""
        found_conflicts = conflicts.get([evc.conflicts.NewDimensionConflict])
        assert len(found_conflicts) == 1
        assert isinstance(found_conflicts[0], evc.conflicts.NewDimensionConflict)

    def test_get_many_types(self, conflicts):
        """Verify that get() fetches many types properly"""
        types = (evc.conflicts.NewDimensionConflict, evc.conflicts.AlgorithmConflict,
                 evc.conflicts.ExperimentNameConflict)
        found_conflicts = conflicts.get(types)
        assert len(found_conflicts) == 3
        assert isinstance(found_conflicts[0], types)
        assert isinstance(found_conflicts[1], types)
        assert isinstance(found_conflicts[2], types)

    def test_get_dimension_name(self, conflicts):
        """Verify that get() fetch the conflict associated to given dimension_name"""
        found_conflicts = conflicts.get(dimension_name='missing')
        assert len(found_conflicts) == 1
        assert isinstance(found_conflicts[0], evc.conflicts.MissingDimensionConflict)
        assert found_conflicts[0].dimension.name == 'missing'

    def test_get_invalid_dimension_name(self, conflicts):
        """Verify that get() raises when dimension_name is invalid"""
        with pytest.raises(ValueError) as exc:
            conflicts.get(dimension_name='invalid')
        assert "Dimension name 'invalid' not found in conflicts" in str(exc.value)

    def test_get_callback(self, conflicts):
        """Verify that get() callback are effectively used"""
        def always_true(conflict):
            return True

        found_conflicts = conflicts.get(callback=always_true)
        assert len(found_conflicts) == 6
        isinstance(found_conflicts[0], evc.conflicts.NewDimensionConflict)
        isinstance(found_conflicts[-1], evc.conflicts.ExperimentNameConflict)

        def always_false(conflict):
            return False

        assert len(conflicts.get(callback=always_false)) == 0

    def test_get_remaining(self, conflicts):
        """Verify that get_remaining() only fetch non resolved conflicts"""
        assert len(conflicts.get_remaining()) == 6
        conflicts.get_remaining()[0].try_resolve()
        assert len(conflicts.get_remaining()) == 5
        conflicts.get_remaining()[0].try_resolve()
        assert len(conflicts.get_remaining()) == 4

    def test_get_resolved(self, conflicts):
        """Verify that get_resolved() only fetch resolved conflicts"""
        assert len(conflicts.get_resolved()) == 0
        conflicts.get_remaining()[0].try_resolve()
        assert len(conflicts.get_resolved()) == 1
        conflicts.get_remaining()[0].try_resolve()
        assert len(conflicts.get_resolved()) == 2

    def test_get_resolutions(self, conflicts, new_dimension_conflict, missing_dimension_conflict):
        """Verify that get_resolution() only fetch unique resolutions"""
        assert len(list(conflicts.get_resolutions())) == 0
        conflicts.try_resolve(missing_dimension_conflict,
                              new_dimension_conflict=new_dimension_conflict)
        assert len(list(conflicts.get_resolutions())) == 1
        assert len(list(conflicts.get_resolved())) == 2
        conflicts.get_remaining()[0].try_resolve()
        assert len(list(conflicts.get_resolutions())) == 2
        assert len(conflicts.get_resolved()) == 3

    def test_are_resolved(self, new_dimension_conflict, missing_dimension_conflict,
                          algorithm_conflict):
        """Verify if resolution status is correct"""
        conflicts = evc.conflicts.Conflicts()
        conflicts.register(new_dimension_conflict)
        conflicts.register(missing_dimension_conflict)
        conflicts.register(algorithm_conflict)

        assert len(conflicts.get_remaining()) == 3
        assert not conflicts.are_resolved
        conflicts.get_remaining()[0].try_resolve()

        assert len(conflicts.get_remaining()) == 2
        assert not conflicts.are_resolved
        conflicts.get_remaining()[0].try_resolve()

        assert len(conflicts.get_remaining()) == 1
        assert not conflicts.are_resolved
        conflicts.get_remaining()[0].try_resolve()

        assert len(conflicts.get_remaining()) == 0
        assert conflicts.are_resolved

    def test_deprecate_existing_conflict(self, new_dimension_conflict, conflicts):
        """Verify deprecated conflicts are not inside conflicts anymore"""
        assert len(conflicts.conflicts) == 6
        conflicts.deprecate([new_dimension_conflict])
        assert len(conflicts.conflicts) == 5

    def test_deprecate_non_existing_conflict(self, conflicts):
        """Verify attempt to deprecate non-existing conflicts raises ValueError"""
        assert len(conflicts.conflicts) == 6
        with pytest.raises(ValueError) as exc:
            conflicts.deprecate(["dummy object"])
        assert "'dummy object' is not in list" in str(exc.value)

    def test_try_resolve_silence_errors(self, capsys, code_conflict, conflicts):
        """Verify try_resolve errors are silenced"""
        conflicts.try_resolve(code_conflict)
        out, err = capsys.readouterr()
        assert (out.split("\n")[-3] ==
                "ValueError: Invalid code change type 'None'. Should be one of "
                "['noeffect', 'break', 'unsure']")

        conflicts.try_resolve(code_conflict, silence_errors=True)
        out, err = capsys.readouterr()
        assert out == ''

    def test_try_resolve_keyboard_interrupt(self, capsys, code_conflict, conflicts):
        """Verify try_resolve catch all errors but not keyboard interrupt"""
        def raise_base_exception(*args, **kwargs):
            raise Exception("Raised properly")
        code_conflict.try_resolve = raise_base_exception
        conflicts.try_resolve(code_conflict)
        out, err = capsys.readouterr()
        assert (out.split("\n")[-3] == "Exception: Raised properly")

        def raise_keyboard_interrupt(*args, **kwargs):
            raise KeyboardInterrupt()
        code_conflict.try_resolve = raise_keyboard_interrupt
        with pytest.raises(KeyboardInterrupt):
            conflicts.try_resolve(code_conflict)

    def test_try_resolve_side_effect_conflicts(
            self, new_dimension_conflict,
            missing_dimension_conflict, conflicts):
        """Verify try_resolve register new conflicts created by resolutions"""
        assert len(conflicts.conflicts) == 6
        conflicts.try_resolve(missing_dimension_conflict,
                              new_dimension_conflict=new_dimension_conflict)
        assert len(conflicts.conflicts) == 7
        assert isinstance(conflicts.conflicts[-1], evc.conflicts.ChangedDimensionConflict)
        assert not conflicts.conflicts[-1].is_resolved



# -*- coding: utf-8 -*-
# pylint:disable=too-many-lines
"""
:mod:`orion.algo.space` -- Objects describing a problem's domain
==================================================================

.. module:: space
   :platform: Unix
   :synopsis: Classes for representing the search space of an
      optimization problem.

There are 3 classes representing possible parameter types. All of them subclass
the base class `Dimension`:

    * `Real`
    * `Integer`
    * `Categorical`

These are instantiated to declare a problem's parameter space. Oríon registers
them in a ordered dictionary, `Space`, which describes how the parameters should
be in order for `orion.algo.base.AbstractAlgorithm` implementations to
communicate with `orion.core`.

Parameter values recorded in `orion.core.worker.trial.Trial` objects must be
and are in concordance with `orion.algo.space` objects. These objects will be
defined by `orion.core` using the user script's configuration file.

Prior distributions, contained in `Dimension` classes, are based on
`scipy.stats.distributions` and should be configured as noted in the
scipy documentation for each specific implentation of a random variable type,
unless noted otherwise!

"""
import copy
import logging
import numbers

import numpy
from scipy.stats import distributions

from orion.core.utils.points import flatten_dims, regroup_dims


logger = logging.getLogger(__name__)


def check_random_state(seed):
    """Return numpy global rng or RandomState if seed is specified"""
    if seed is None or seed is numpy.random:
        rng = numpy.random.mtrand._rand  # pylint:disable=protected-access,c-extension-no-member
    elif isinstance(seed, numpy.random.RandomState):
        rng = seed
    else:
        try:
            rng = numpy.random.RandomState(seed)
        except Exception as e:
            raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                             ' instance' % seed) from e

    return rng


# helper class to be able to print [1, ..., 4] instead of [1, '...', 4]
class _Ellipsis:  # pylint:disable=too-few-public-methods
    def __repr__(self):
        return '...'


class Dimension:
    """Base class for search space dimensions.

    Attributes
    ----------
    name : str
       Unique identifier for this `Dimension`.
    type : str
       Identifier for the type of parameters this `Dimension` is representing.
       it can be 'real', 'integer', or 'categorical' (name of a subclass).
    prior : `scipy.stats.distributions.rv_generic`
       A distribution over the original dimension.
    shape : tuple
       Defines how many dimensions are packed in this `Dimension`.
       Describes the shape of the corresponding tensor.

    """

    NO_DEFAULT_VALUE = None

    def __init__(self, name, prior, *args, **kwargs):
        """Init code which is common for `Dimension` subclasses.

        Parameters
        ----------
        name : str
           Unique identifier associated with this `Dimension`,
           e.g. 'learning_rate'.
        prior : str | `scipy.stats.distributions.rv_generic`
           Corresponds to a name of an instance or an instance itself of
           `scipy.stats._distn_infrastructure.rv_generic`. Basically,
           the name of the distribution one wants to use as a :attr:`prior`.
        args : list
        kwargs : dict
           Shape parameter(s) for the `prior` distribution.
           Should include all the non-optional arguments.
           It may include ``loc``, ``scale``, ``shape``.

        .. seealso:: `scipy.stats.distributions` for possible values of
           `prior` and their arguments.

        """
        self._name = None
        self.name = name

        if isinstance(prior, str):
            self._prior_name = prior
            self.prior = getattr(distributions, prior)
        elif prior is None:
            self._prior_name = "None"
            self.prior = prior
        else:
            self._prior_name = prior.name
            self.prior = prior
        self._args = args
        self._kwargs = kwargs
        self._default_value = kwargs.pop('default_value', self.NO_DEFAULT_VALUE)
        self._shape = kwargs.pop('shape', None)
        self.validate()

    def validate(self):
        """Validate dimension arguments"""
        if 'random_state' in self._kwargs or 'seed' in self._kwargs:
            raise ValueError("random_state/seed cannot be set in a "
                             "parameter's definition! Set seed globally!")
        if 'discrete' in self._kwargs:
            raise ValueError("Do not use kwarg 'discrete' on `Dimension`, "
                             "use pure `_Discrete` class instead!")
        if 'size' in self._kwargs:
            raise ValueError("Use 'shape' keyword only instead of 'size'.")

        if self.default_value is not self.NO_DEFAULT_VALUE and self.default_value not in self:
            raise ValueError("{} is not a valid value for this Dimension. "
                             "Can't set default value.".format(self.default_value))

    def _get_hashable_members(self):
        return (self.name, self.shape, self.type, tuple(self._args), tuple(self._kwargs.items()),
                self.default_value, self._prior_name)

    # pylint:disable=protected-access
    def __eq__(self, other):
        """Return True if other is the same dimension as self"""
        if not isinstance(other, Dimension):
            return False

        return self._get_hashable_members() == other._get_hashable_members()

    def __hash__(self):
        """Return the hash of the hashable members"""
        return hash(self._get_hashable_members())

    def sample(self, n_samples=1, seed=None):
        """Draw random samples from `prior`.

        Parameters
        ----------
        n_samples : int, optional
           The number of samples to be drawn. Default is 1 sample.
        seed : None | int | ``numpy.random.RandomState`` instance, optional
           This parameter defines the RandomState object to use for drawing
           random variates. If None (or np.random), the **global**
           np.random state is used. If integer, it is used to seed a
           RandomState instance **just for the call of this function**.
           Default is None.

           Set random state to something other than None for reproducible
           results.

        .. warning:: Setting `seed` with an integer will cause the same ndarray
           to be sampled if ``n_samples > 0``. Set `seed` with a
           ``numpy.random.RandomState`` to carry on the changes in random state
           across many samples.

        """
        samples = [self.prior.rvs(*self._args, size=self.shape,
                                  random_state=seed,
                                  **self._kwargs) for _ in range(n_samples)]
        return samples

    def cast(self, point):
        """Cast a point to dimension's type

        If casted point will stay a list or a numpy array depending on the
        given point's type.
        """
        raise NotImplementedError

    def interval(self, alpha=1.0):
        """Return a tuple containing lower and upper bound for parameters.

        If parameters are drawn from an 'open' supported random variable,
        then it will be attempted to calculate the interval from which
        a variable is `alpha`-likely to be drawn from.

        """
        return self.prior.interval(alpha, *self._args, **self._kwargs)

    def __contains__(self, point):
        """Check if constraints hold for this `point` of `Dimension`.

        :param point: a parameter corresponding to this `Dimension`.
        :type point: numeric or array-like

        .. note:: Default `Dimension` does not have any extra constraints.
           It just checks whether point lies inside the support and the shape.

        """
        raise NotImplementedError

    def __repr__(self):
        """Represent the object as a string."""
        return "{0}(name={1}, prior={{{2}: {3}, {4}}}, shape={5}, default value={6})".format(
            self.__class__.__name__, self.name, self._prior_name,
            self._args, self._kwargs, self.shape, self._default_value)

    def get_prior_string(self):
        """Build the string corresponding to current prior"""
        args = copy.deepcopy(list(self._args[:]))
        if self._prior_name == 'uniform' and len(args) == 2:
            args[1] = args[0] + args[1]
            args[0] = args[0]

        args = list(map(str, args))

        for k, v in self._kwargs.items():
            if isinstance(v, str):
                args += ["{}='{}'".format(k, v)]
            else:
                args += ["{}={}".format(k, v)]

        if self._shape is not None:
            args += ['shape={}'.format(self._shape)]
        if self.default_value is not self.NO_DEFAULT_VALUE:
            args += ['default_value={}'.format(repr(self.default_value))]

        prior_name = self._prior_name
        if prior_name == 'reciprocal':
            prior_name = 'loguniform'
        return "{prior_name}({args})".format(prior_name=prior_name, args=", ".join(args))

    def get_string(self):
        """Build the string corresponding to current dimension"""
        return "{name}~{prior}".format(name=self.name, prior=self.get_prior_string())

    @property
    def name(self):
        """See `Dimension` attributes."""
        return self._name

    @name.setter
    def name(self, value):
        if isinstance(value, str) or value is None:
            self._name = value
        else:
            raise TypeError("Dimension's name must be either string or None. "
                            "Provided: {}, of type: {}".format(value, type(value)))

    @property
    def default_value(self):
        """Return the default value for this dimensions"""
        return self._default_value

    @property
    def type(self):
        """See `Dimension` attributes."""
        return self.__class__.__name__.lower()

    @property
    def prior_name(self):
        """Return the name of the prior"""
        return self._prior_name

    @property
    def shape(self):
        """Return the shape of dimension."""
        # Default shape `None` corresponds to 0-dim (scalar) or shape == ().
        # Read about ``size`` argument in
        # `scipy.stats._distn_infrastructure.rv_generic._argcheck_rvs`
        if self.prior is None:
            return None

        _, _, _, size = self.prior._parse_args_rvs(*self._args,  # pylint:disable=protected-access
                                                   size=self._shape,
                                                   **self._kwargs)
        return size

    # pylint:disable=no-self-use
    @property
    def cardinality(self):
        """Return the number of all the possible points from `Dimension`.
        The default value is `numpy.inf`.
        """
        return numpy.inf


def _is_numeric_array(point):
    """Test whether a point is numerical object or an array containing only numerical objects"""
    def _is_numeric(item):
        return isinstance(item, (numbers.Number, numpy.ndarray))

    try:
        return all(_is_numeric(item) for item in point)
    except TypeError:
        return _is_numeric(point)

    return False


class Real(Dimension):
    """Search space dimension that can take on any real value.

    Parameters
    ----------
    name : str
    prior : str
       See Parameters of `Dimension.__init__`.
    args : list
    kwargs : dict
       See Parameters of `Dimension.__init__` for general.

    Real kwargs (extra)
    -------------------
    low : float
       Lower bound (inclusive), optional; default ``-numpy.inf``.
    high : float:
       Upper bound (inclusive), optional; default ``numpy.inf``.
       The upper bound must be inclusive because of rounding errors
       during optimization which may cause values to round exactly
       to the upper bound.
    precision : int
        Precision, optional; default ``4``.
    shape : tuple
       Defines how many dimensions are packed in this `Dimension`.
       Describes the shape of the corresponding tensor.

    """

    def __init__(self, name, prior, *args, **kwargs):
        self._low = kwargs.pop('low', -numpy.inf)
        self._high = kwargs.pop('high', numpy.inf)
        if self._high <= self._low:
            raise ValueError("Lower bound {} has to be less than upper bound {}"
                             .format(self._low, self._high))

        precision = kwargs.pop('precision', 4)
        if (isinstance(precision, int) and precision > 0) or precision is None:
            self.precision = precision
        else:
            raise TypeError("Precision should be a non-negative int or None, "
                            "instead was {} of type {}.".format(precision, type(precision)))

        super(Real, self).__init__(name, prior, *args, **kwargs)

    def __contains__(self, point):
        """Check if constraints hold for this `point` of `Dimension`.

        :param point: a parameter corresponding to this `Dimension`.
        :type point: numeric or array-like

        .. note:: Default `Dimension` does not have any extra constraints.
           It just checks whether point lies inside the support and the shape.

        """
        if not _is_numeric_array(point):
            return False

        low, high = self.interval()

        point_ = numpy.asarray(point)

        if point_.shape != self.shape:
            return False

        return numpy.all(point_ >= low) and numpy.all(point_ <= high)

    def interval(self, alpha=1.0):
        """Return a tuple containing lower and upper bound for parameters.

        If parameters are drawn from an 'open' supported random variable,
        then it will be attempted to calculate the interval from which
        a variable is `alpha`-likely to be drawn from.

        .. note:: Both lower and upper bounds are inclusive.

        """
        prior_low, prior_high = super(Real, self).interval(alpha)
        return (max(prior_low, self._low), min(prior_high, self._high))

    def sample(self, n_samples=1, seed=None):
        """Draw random samples from `prior`.

        .. seealso:: `Dimension.sample`

        """
        samples = []
        for _ in range(n_samples):
            for _ in range(4):
                sample = super(Real, self).sample(1, seed)
                if sample[0] not in self:
                    nice = False
                    continue
                nice = True
                samples.extend(sample)
                break
            if not nice:
                raise ValueError("Improbable bounds: (low={0}, high={1}). "
                                 "Please make interval larger.".format(self._low, self._high))

        return samples

    # pylint:disable=no-self-use
    def cast(self, point):
        """Cast a point to float

        If casted point will stay a list or a numpy array depending on the
        given point's type.
        """
        casted_point = numpy.asarray(point).astype(float)

        if not isinstance(point, numpy.ndarray):
            return casted_point.tolist()

        return casted_point


class _Discrete(Dimension):

    def sample(self, n_samples=1, seed=None):
        """Draw random samples from `prior`.

        Discretizes with `numpy.floor` the results from `Dimension.sample`.

        .. seealso:: `Dimension.sample`
        .. seealso:: Discussion in https://github.com/epistimio/orion/issues/56
           if you want to understand better how this `Integer` diamond inheritance
           works.

        """
        samples = super(_Discrete, self).sample(n_samples, seed)
        # Making discrete by ourselves because scipy does not use **floor**
        return list(map(lambda x: numpy.floor(x).astype(int), samples))

    def interval(self, alpha=1.0):
        """Return a tuple containing lower and upper bound for parameters.

        If parameters are drawn from an 'open' supported random variable,
        then it will be attempted to calculate the interval from which
        a variable is `alpha`-likely to be drawn from.

        Bounds are integers.

        .. note:: Both lower and upper bounds are inclusive.

        """
        low, high = super(_Discrete, self).interval(alpha)
        try:
            int_low = int(numpy.floor(low))
        except OverflowError:  # infinity cannot be converted to Python int type
            int_low = -numpy.inf
        try:
            int_high = int(numpy.ceil(high))
        except OverflowError:  # infinity cannot be converted to Python int type
            int_high = numpy.inf
        return (int_low, int_high)

    def __contains__(self, point):
        raise NotImplementedError


class Integer(Real, _Discrete):
    """Search space dimension representing integer values.

    Parameters
    ----------
    name : str
    prior : str
       See Parameters of `Dimension.__init__`.
    args : list
    kwargs : dict
       See Parameters of `Dimension.__init__` for general.

    Real kwargs (extra)
    -------------------
    low : float
       Lower bound (inclusive), optional; default ``-numpy.inf``.
    high : float:
       Upper bound (inclusive), optional; default ``numpy.inf``.
    precision : int
        Precision, optional; default ``4``.
    shape : tuple
       Defines how many dimensions are packed in this `Dimension`.
       Describes the shape of the corresponding tensor.

    """

    def __contains__(self, point):
        """Check if constraints hold for this `point` of `Dimension`.

        :param point: a parameter corresponding to this `Dimension`.
        :type point: numeric or array-like

        `Integer` will check whether `point` contains only integers.

        """
        if not _is_numeric_array(point):
            return False

        point_ = numpy.asarray(point)
        if not numpy.all(numpy.equal(numpy.mod(point_, 1), 0)):
            return False

        return super(Integer, self).__contains__(point)

    # pylint:disable=no-self-use
    def cast(self, point):
        """Cast a point to int

        If casted point will stay a list or a numpy array depending on the
        given point's type.
        """
        casted_point = numpy.asarray(point).astype(int)

        if not isinstance(point, numpy.ndarray):
            return casted_point.tolist()

        return casted_point

    def get_prior_string(self):
        """Build the string corresponding to current prior"""
        prior_string = super(Integer, self).get_prior_string()
        return prior_string[:-1] + ', discrete=True)'

    @property
    def prior_name(self):
        """Return the name of the prior"""
        return 'int_{}'.format(super(Integer, self).prior_name)

    @property
    def cardinality(self):
        """Return the number of all the possible points from Integer `Dimension`"""
        low, high = self.interval()
        return _get_shape_cardinality(self.shape) * int(high - low)


def _get_shape_cardinality(shape):
    """Get the cardinality in a shape which can be int or tuple"""
    shape_cardinality = 1
    if shape is None:
        return shape_cardinality

    if isinstance(shape, int):
        shape = (shape, )

    for cardinality in shape:
        shape_cardinality *= cardinality
    return shape_cardinality


class Categorical(Dimension):
    """Search space dimension that can take on categorical values.

    Parameters
    ----------
    name : str
       See Parameters of `Dimension.__init__`.
    categories : dict or other iterable
       A dictionary would associate categories to probabilities, else
       it assumes to be drawn uniformly from the iterable.
    kwargs : dict
       See Parameters of `Dimension.__init__` for general.

    """

    def __init__(self, name, categories, **kwargs):
        if isinstance(categories, dict):
            self.categories = tuple(categories.keys())
            self._probs = tuple(categories.values())
        else:
            self.categories = tuple(categories)
            self._probs = tuple(numpy.tile(1. / len(categories), len(categories)))

        # Just for compatibility; everything should be `Dimension` to let the
        # `Transformer` decorators be able to wrap smoothly anything.
        prior = distributions.rv_discrete(values=(list(range(len(self.categories))),
                                                  self._probs))
        super(Categorical, self).__init__(name, prior, **kwargs)

    @property
    def cardinality(self):
        """Return the number of all the possible values from Categorical `Dimension`"""
        return len(self.categories) * _get_shape_cardinality(self._shape)

    def sample(self, n_samples=1, seed=None):
        """Draw random samples from `prior`.

        .. seealso:: `Dimension.sample`

        """
        rng = check_random_state(seed)
        cat_ndarray = numpy.array(self.categories, dtype=numpy.object)
        samples = [rng.choice(cat_ndarray, p=self._probs, size=self._shape)
                   for _ in range(n_samples)]
        return samples

    def interval(self, alpha=1.0):
        """Return a tuple of possible values that this categorical dimension can take."""
        return self.categories

    def __contains__(self, point):
        """Check if constraints hold for this `point` of `Dimension`.

        :param point: a parameter corresponding to this `Dimension`.
        :type point: numeric or array-like

        """
        point_ = numpy.asarray(point, dtype=numpy.object)
        if point_.shape != self.shape:
            return False
        _check = numpy.vectorize(lambda x: x in self.categories)
        return numpy.all(_check(point_))

    def __repr__(self):
        """Represent the object as a string."""
        if len(self.categories) > 5:
            cats = self.categories[:2] + self.categories[-2:]
            probs = self._probs[:2] + self._probs[-2:]
            prior = list(zip(cats, probs))
            prior.insert(2, _Ellipsis())
        else:
            cats = self.categories
            probs = self._probs
            prior = list(zip(cats, probs))

        prior = map(lambda x: '{0[0]}: {0[1]:.2f}'.format(x)
                    if not isinstance(x, _Ellipsis) else str(x), prior)

        prior = "{" + ', '.join(prior) + "}"

        return "Categorical(name={0}, prior={1}, shape={2}, default value={3})"\
               .format(self.name, prior, self.shape, self.default_value)

    def get_prior_string(self):
        """Build the string corresponding to current prior"""
        args = list(map(str, self._args[:]))
        args += ["{}={}".format(k, v) for k, v in self._kwargs.items()]
        if self.default_value is not self.NO_DEFAULT_VALUE:
            args += ['default_value={}'.format(self.default_value)]

        cats = [repr(c) for c in self.categories]
        if all(p == self._probs[0] for p in self._probs):
            prior = '[{}]'.format(", ".join(cats))
        else:
            probs = list(zip(cats, self._probs))
            prior = '{' + ", ".join('{0}: {1:.2f}'.format(c, p) for c, p in probs) + '}'

        args = [prior]

        if self.default_value is not self.NO_DEFAULT_VALUE:
            args += ['default_value={}'.format(repr(self.default_value))]

        return 'choices({args})'.format(args=', '.join(args))

    @property
    def get_prior(self):
        """Return the priors"""
        return self._probs

    @property
    def prior_name(self):
        """Return the name of the prior"""
        return "choices"

    def cast(self, point):
        """Cast a point to some category

        Casted point will stay a list or a numpy array depending on the
        given point's type.

        Raises
        ------
        ValueError
            If one of the category in `point` is not present in current Categorical Dimension.

        """
        categorical_strings = {str(c): c for c in self.categories}

        def get_category(value):
            """Return category corresponding to a string else return singleton object"""
            if str(value) not in categorical_strings:
                raise ValueError("Invalid category: {}".format(value))

            return categorical_strings[str(value)]

        point_ = numpy.asarray(point, dtype=numpy.object)
        cast = numpy.vectorize(get_category, otypes=[numpy.object])
        casted_point = cast(point_)

        if not isinstance(point, numpy.ndarray):
            return casted_point.tolist()

        return casted_point


class Fidelity(Dimension):
    """Fidelity `Dimension` for representing multi-fidelity.

    Fidelity dimensions are not optimized by the algorithms. If it supports multi-fidelity, the
    algorithm will select a fidelity level for which it will sample hyper-parameter values to
    explore a low fidelity space. This class is used as a place-holder so that algorithms can
    discern fidelity dimensions from hyper-parameter dimensions.

    Parameters
    ----------
    name : str
        Name of the dimension
    low: int
        Mininum of the fidelity interval.
    high: int
        Maximum of the fidelity interval.
    base: int
        Base logarithm of the fidelity dimension.

    Attributes
    ----------
    name : str
        Name of the dimension
    default_value: int
        Maximum of the fidelity interval.

    """

    # pylint:disable=super-init-not-called
    def __init__(self, name, low, high, base=2):
        if low <= 0:
            raise AttributeError("Minimum resources must be a positive number.")
        elif low > high:
            raise AttributeError("Minimum resources must be smaller than maximum resources.")
        if base < 1:
            raise AttributeError("Base should be greater than or equal to 1")
        self.name = name
        self.low = int(low)
        self.high = int(high)
        self.base = int(base)
        self.prior = None
        self._prior_name = 'None'

    @property
    def default_value(self):
        """Return `high`"""
        return self.high

    # pylint:disable=no-self-use
    @property
    def cardinality(self):
        """Return cardinality of Fidelity dimension, leave it to 1 as Fidelity dimension
        does not contribute to cardinality in a fixed way now.
        """
        return 1

    def get_prior_string(self):
        """Build the string corresponding to current prior"""
        return 'fidelity({}, {}, {})'.format(self.low, self.high, self.base)

    def validate(self):
        """Do not do anything."""
        raise NotImplementedError

    def sample(self, n_samples=1, seed=None):
        """Do not do anything."""
        return [self.high for i in range(n_samples)]

    def interval(self, alpha=1.0):
        """Do not do anything."""
        return (self.low, self.high)

    def cast(self, point=0):
        """Do not do anything."""
        raise NotImplementedError

    def __repr__(self):
        """Represent the object as a string."""
        return "{0}(name={1}, low={2}, high={3}, base={4})".format(
            self.__class__.__name__, self.name, self.low, self.high, self.base)

    def __contains__(self, value):
        """Check if constraints hold for this `point` of `Dimension`.

        :param point: a parameter corresponding to this `Dimension`.
        :type point: numeric or array-like
        """
        return self.low <= value <= self.high


class Space(dict):
    """Represents the search space.

    It is a sorted dictionary which contains `Dimension` objects.
    The dimensions are sorted based on their names.
    """

    contains = Dimension

    def register(self, dimension):
        """Register a new dimension to `Space`."""
        self[dimension.name] = dimension

    def sample(self, n_samples=1, seed=None):
        """Draw random samples from this space.

        Parameters
        ----------
        n_samples : int, optional
           The number of samples to be drawn. Default is 1 sample.
        seed : None | int | ``numpy.random.RandomState`` instance, optional
           This parameter defines the RandomState object to use for drawing
           random variates. If None (or np.random), the **global**
           np.random state is used. If integer, it is used to seed a
           RandomState instance **just for the call of this function**.
           Default is None.

           Set random state to something other than None for reproducible
           results.

        Returns
        -------
        points : list of tuples of array-likes
           Each element is a separate sample of this space, a list containing
           values associated with the corresponding dimension. Values are in the
           same order as the contained dimensions. Their shape is determined
           by ``dimension.shape``.

        """
        rng = check_random_state(seed)
        samples = [dim.sample(n_samples, rng) for dim in self.values()]
        return list(zip(*samples))

    def interval(self, alpha=1.0):
        """Return a list with the intervals for each contained dimension."""
        res = list()
        for dim in self.values():
            if dim.type == 'categorical':
                res.append(dim.categories)
            else:
                res.append(dim.interval(alpha))
        return res

    def __getitem__(self, key):
        """Wrap __getitem__ to allow searching with position."""
        if isinstance(key, str):
            return super(Space, self).__getitem__(key)

        values = list(self.values())
        return values[key]

    def __setitem__(self, key, value):
        """Wrap __setitem__ to allow only ``Space.contains`` class, e.g. `Dimension`,
        values and string keys.
        """
        if not isinstance(key, str):
            raise TypeError("Keys registered to {} must be string types. "
                            "Provided: {}".format(self.__class__.__name__, key))
        if not isinstance(value, self.contains):
            raise TypeError("Values registered to {} must be {} types. "
                            "Provided: {}".format(self.__class__.__name__,
                                                  self.contains.__name__, value))
        if key in self:
            raise ValueError("There is already a Dimension registered with this name. "
                             "Register it with another name. Provided: {}".format(key))
        super(Space, self).__setitem__(key, value)

    def __contains__(self, value):
        """Check whether `value` is within the bounds of the space.
        Or check if a name for a dimension is registered in this space.

        :param value: list of values associated with the dimensions contained
           or a string indicating a dimension's name.

        """
        if isinstance(value, str):
            return super(Space, self).__contains__(value)

        try:
            len(value)
        except TypeError as exc:
            raise TypeError("Can check only for dimension names or "
                            "for tuples with parameter values.") from exc

        if not self:
            return False

        for component, dim in zip(value, self.values()):
            if component not in dim:
                return False

        return True

    def __repr__(self):
        """Represent as a string the space and the dimensions it contains."""
        dims = list(self.values())
        return "Space([{}])".format(',\n       '.join(map(str, dims)))

    def items(self):
        """Return items sorted according to keys"""
        return [(k, self[k]) for k in self.keys()]

    def values(self):
        """Return values sorted according to keys"""
        return [self[k] for k in self.keys()]

    def keys(self):
        """Return sorted keys"""
        return list(iter(self))

    def __iter__(self):
        """Return sorted keys"""
        return iter(sorted(super(Space, self).keys()))

    @property
    def configuration(self):
        """Return a dictionary of priors."""
        return {name: dim.get_prior_string() for name, dim in self.items()}

    @property
    def cardinality(self):
        """Return the number of all all possible sets of samples in the space"""
        capacities = 1
        for dim in self.values():
            capacities *= dim.cardinality
        return capacities


def pack_point(point, space):
    """Take a list of points and pack it appropriately as a point from `space`.

    This function is deprecated and will be removed in v0.2.0. Use
    `orion.core.utils.points.regroup_dims` instead.
    """
    logger.warning('`pack_point` is deprecated and will be removed in v0.2.0. Use '
                   '`orion.core.utils.points.regroup_dims` instead.')
    return regroup_dims(point, space)


def unpack_point(point, space):
    """Flatten `point` in `space` and convert it to a 1D `numpy.ndarray`.

    This function is deprecated and will be removed in v0.2.0. Use
    `orion.core.utils.points.flatten_dims` instead.
    """
    logger.warning('`unpack_point` is deprecated and will be removed in v0.2.0. Use '
                   '`orion.core.utils.points.regroup_dims` instead.')
    return flatten_dims(point, space)



import sys

import numpy as np
from orion.client import report_objective
from sklearn.datasets import load_iris
from sklearn.linear_model import SGDClassifier
from sklearn.metrics import balanced_accuracy_score
from sklearn.model_selection import train_test_split

# Make the execution reproducible
np.random.seed(1)

# Parsing the value for the hyper-parameter 'epsilon' given as a command line argument.
hyper_epsilon = sys.argv[1]
print("Epsilon is {}".format(hyper_epsilon))

# Loading the iris dataset and splitting it into training and testing set.
X, y = load_iris(return_X_y=True)
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3)

# Training the model with the training set with the specified 'epsilon' to control the huber loss.
clf = SGDClassifier(loss='huber', epsilon=float(hyper_epsilon))
clf.fit(X_train, y_train)

# Evaluating the accuracy using the testing set.
y_pred = clf.predict(X_test)
accuracy = balanced_accuracy_score(y_test, y_pred)

# Reporting the results
print("Accuracy is {}".format(accuracy))

report_objective(1 - accuracy)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test for client helper functions."""
import os

import pytest

import orion.core.cli
from orion.core.worker.consumer import Consumer


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_interrupt(database, monkeypatch, capsys):
    """Test interruption from within user script."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "-x~uniform(-50, 50, precision=5)"]

    error_code = orion.core.cli.main([
        "hunt", "--config", "./orion_config.yaml", "--worker-trials", "2",
        "python", "black_box.py", "interrupt_trial"] + user_args)

    assert error_code == 130

    captured = capsys.readouterr()
    assert captured.out == 'Orion is interrupted.\n'
    assert captured.err == ''

    exp = list(database.experiments.find({'name': 'voila_voici'}))
    exp = exp[0]
    exp_id = exp['_id']
    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) == 1
    assert trials[0]['status'] == 'interrupted'


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_interrupt_diff_code(database, monkeypatch, capsys):
    """Test interruption from within user script with custom int code"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "-x~uniform(-50, 50, precision=5)"]

    # Set local to 200
    orion.core.config.worker.interrupt_signal_code = 200

    # But child won't be passed ORION_INTERRUPT_CODE and therefore will send default code 130
    def empty_env(self, trial, results_file=None):
        return os.environ

    with monkeypatch.context() as m:
        m.setattr(Consumer, 'get_execution_environment', empty_env)

        # Interrupt won't be interpreted properly and trials will be marked as broken
        error_code = orion.core.cli.main([
            "hunt", "--config", "./orion_config.yaml", "--worker-trials", "2",
            "python", "black_box.py", "interrupt_trial"] + user_args)

        assert error_code == 0

        exp = list(database.experiments.find({'name': 'voila_voici'}))
        exp = exp[0]
        exp_id = exp['_id']
        trials = list(database.trials.find({'experiment': exp_id}))
        assert len(trials) == 2
        assert trials[0]['status'] == 'broken'

    # This time we use true `get_execution_environment which pass properly int code to child.
    error_code = orion.core.cli.main([
        "hunt", "--config", "./orion_config.yaml", "--worker-trials", "2",
        "python", "black_box.py", "interrupt_trial"] + user_args)

    assert error_code == 130

    captured = capsys.readouterr()
    assert 'Orion is interrupted.\n' in captured.out
    assert captured.err == ''

    exp = list(database.experiments.find({'name': 'voila_voici'}))
    exp = exp[0]
    exp_id = exp['_id']
    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) == 3
    assert trials[-1]['status'] == 'interrupted'


# TODO:

# test no call to any report

# Add all this in DOC.


@pytest.mark.parametrize('fct', ['report_bad_trial', 'report_objective'])
@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_report_no_name(database, monkeypatch, fct):
    """Test report helper functions with default names"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main([
        "hunt", "--config", "./orion_config.yaml", "--worker-trials", "2",
        "python", "black_box.py", fct, "--objective", "1.0"] + user_args)

    exp = list(database.experiments.find({'name': 'voila_voici'}))
    exp = exp[0]
    exp_id = exp['_id']
    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) == 2
    assert trials[0]['status'] == 'completed'
    assert trials[0]['results'][0]['name'] == 'objective'
    assert trials[0]['results'][0]['type'] == 'objective'
    assert trials[0]['results'][0]['value'] == 1.0


@pytest.mark.parametrize('fct', ['report_bad_trial', 'report_objective'])
@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_report_with_name(database, monkeypatch, fct):
    """Test report helper functions with custom names"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main([
        "hunt", "--config", "./orion_config.yaml", "--worker-trials", "2",
        "python", "black_box.py", fct, "--objective", "1.0", "--name", "metric"] + user_args)

    exp = list(database.experiments.find({'name': 'voila_voici'}))
    exp = exp[0]
    exp_id = exp['_id']
    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) == 2
    assert trials[0]['status'] == 'completed'
    assert trials[0]['results'][0]['name'] == 'metric'
    assert trials[0]['results'][0]['type'] == 'objective'
    assert trials[0]['results'][0]['value'] == 1.0


@pytest.mark.parametrize('fct', ['report_bad_trial', 'report_objective'])
@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_report_with_bad_objective(database, monkeypatch, fct):
    """Test report helper functions with bad objective types"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "-x~uniform(-50, 50, precision=5)"]

    with pytest.raises(ValueError) as exc:
        orion.core.cli.main([
            "hunt", "--config", "./orion_config.yaml", "--worker-trials", "2",
            "python", "black_box.py", fct, "--objective", "oh oh"] + user_args)

    assert 'must contain a type `objective` with type float/int' in str(exc.value)


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_report_with_bad_trial_no_objective(database, monkeypatch):
    """Test bad trial report helper function with default objective."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main([
        "hunt", "--config", "./orion_config.yaml", "--worker-trials", "2",
        "python", "black_box.py", "report_bad_trial"] + user_args)

    exp = list(database.experiments.find({'name': 'voila_voici'}))
    exp = exp[0]
    exp_id = exp['_id']
    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) == 2
    assert trials[0]['status'] == 'completed'
    assert trials[0]['results'][0]['name'] == 'objective'
    assert trials[0]['results'][0]['type'] == 'objective'
    assert trials[0]['results'][0]['value'] == 1e10


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_report_with_bad_trial_with_data(database, monkeypatch):
    """Test bad trial report helper function with additional data."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "-x~uniform(-50, 50, precision=5)"]

    orion.core.cli.main([
        "hunt", "--config", "./orion_config.yaml", "--worker-trials", "2",
        "python", "black_box.py", "report_bad_trial", "--data", "another"] + user_args)

    exp = list(database.experiments.find({'name': 'voila_voici'}))
    exp = exp[0]
    exp_id = exp['_id']
    trials = list(database.trials.find({'experiment': exp_id}))
    assert len(trials) == 2
    assert trials[0]['status'] == 'completed'
    assert trials[0]['results'][0]['name'] == 'objective'
    assert trials[0]['results'][0]['type'] == 'objective'
    assert trials[0]['results'][0]['value'] == 1e10

    assert trials[0]['results'][1]['name'] == 'another'
    assert trials[0]['results'][1]['type'] == 'constraint'
    assert trials[0]['results'][1]['value'] == 1.0


@pytest.mark.usefixtures("clean_db")
@pytest.mark.usefixtures("null_db_instances")
def test_no_report(database, monkeypatch, capsys):
    """Test script call without any results reported."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))

    user_args = [
        "-x~uniform(-50, 50, precision=5)"]

    errorcode = orion.core.cli.main([
        "hunt", "--config", "./orion_config.yaml", "--worker-trials", "2",
        "python", "black_box.py", 'no_report'] + user_args)

    assert errorcode == 1

    captured = capsys.readouterr()
    assert captured.out == ""
    assert 'Cannot parse result file' in captured.err



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.io.cmdline_parser` -- Parse command line arguments for Orion
=============================================================================

.. module:: cmdline_parser
   :platform: Unix
   :synopsis: Parsing and building of the command line input for using script

Simplify the parsing of a command line by storing every values inside a `dict`
mapping the name of the argument to its value as a key-pair relation. Positional arguments
are stored in the format `_X` where `X` represent the index of that argument inside
the command line.

CmdlineParser provides an interface to parse command line arguments from input but also
templates to build it again as a list or an already formatted string.
"""
from collections import OrderedDict
import os


class CmdlineParser(object):
    """Simple class for commandline arguments parsing.

    `CmdlineParser` provides a simple interface to interpret commandline arguments
    and create new commandlines. This class exposes an interface to transform a list of strings
    representing the commandline in a dictionary of key-value conserving the order of
    the arguments, the value of named arguments, be it boolean, single-valued or multiple ones,
    as well as handling the definition of argument through the `=` sign.

    Attributes
    ----------
    arguments : dict
        Commandline arguments' name and value(s).
    template : list
        List of template-ready strings for formatting.

    See Also
    --------
    parse : Parse the list of string and defines the form of the template.

    Notes
    -----
    Subcommands are not supported and will be likely lead to bad interpretation of the commandline.
    Aggregation of single characters arguments is not supported yet. Ex: `-xzvf`

    """

    def __init__(self):
        """See `CmdlineParser` description"""
        # TODO Handle parsing twice.
        self.keys = dict()
        self.arguments = OrderedDict()
        self._already_parsed = False
        self.template = []

    def get_state_dict(self):
        """Give state dict that can be used to reconstruct the parser"""
        return dict(
            arguments=list(map(list, self.arguments.items())),
            keys=list(map(list, self.keys.items())),
            template=self.template)

    def set_state_dict(self, state):
        """Reset the parser based on previous state"""
        if state.get('keys') is None:
            # NOTE: To support experiments prior to 0.1.9
            state['keys'] = [(key, self._key_to_arg(key)) for key in state['arguments']]
        self.keys = OrderedDict(state['keys'])
        self.arguments = OrderedDict(state['arguments'])
        self.template = state['template']
        self._already_parsed = bool(self.template)

    def format(self, configuration):
        """Format the current template.

        Return a list of string where the arguments have been given the values inside
        the `configuration` argument.

        Parameters
        ----------
        configuration : dict
            Dictionary containing the keys and values for each argument of the commandline.

        Returns
        -------
        list
            A list ressembling the one given to the `parse` method but with values of
            `configuration`.

        """
        # Arguments with spaces are broken if we use ' '.join(template).format().split(' ').
        # Hence we iterate over the list as-is and format on each item separately.
        formatted = []
        for item in self.template:
            if item.startswith('-'):
                formatted.append(item)
            else:
                formatted.append(item.format(**configuration))

        return formatted

    def parse(self, commandline):
        """Parse the `commandline` argument.

        This function will go through the list of strings and create a dictionary
        mapping every argument to its value(s). It will also form a template to recreate such a
        list. If `commandline` is empty or the `parse` function has already been called, it
        will return the dictionary of the already-parsed arguments.

        Parameters
        ----------
        commandline : list
            List of string representing the commmandline arguments.

        Returns
        -------
        dict
            Dictionary holding the values of every argument. The keys are the name of the arguments.

        Raises
        ------
        ValueError
            This exception is raised then the parser detects a duplicate argument.
        RuntimeError
            This exception is raised if the parser already parsed a commandline and contains a
            template.

        Notes
        -----
        Each argument is stored inside a `dict`.
        Inside that dictionary, the keys are created following these rules:

        -If the argument is a positional one, its key will be `'_pos_x'` where `x` is its index
        inside the list.
        -Otherwise, the key will be equal to the name of the argument minus the prefixed dashes.

        The values are stored in the following ways:

        -If the argument is a positional one, its value is simply its value inside the commandline.
        -If the argument is a named, boolean argument, its value is simply `True`.
        -If the argument is a named, single-valued argument, its value is stored as-is.
        -If the argument is a named, multi-valued argument, its value is a list containing each
        value following it until the next named argument.

        Positional arguments following a named argument are not currently supported.

        Examples
        --------
        Positional and named arguments:

        >>> parser = CmdlineParser()
        >>> parser.parse('python 1 --arg value'.split(' '))
        {'_pos_0': 'python', '_pos_1': '1', 'arg': 'value'}

        Named boolean argument:

        >>> parser.parse('python --boolean'.split(' '))
        {'_pos_0': 'python', 'boolean': True}

        Named multi-valued argument:

        >>> parser.parse('python --args value1 value2'.split(' '))
        {'_pos_0': 'python', 'args': ['value1', 'value2']}

        Named argument defined with `=`:

        >>> parser.parse('python --arg=value'.split(' '))
        {'_pos_0': 'python', 'arg': 'value'}

        """
        if self._already_parsed:
            raise RuntimeError("The commandline has already been parsed.")

        keys, arguments = self._parse_arguments(commandline)
        self.arguments = arguments
        self.keys = keys

        for key, value in arguments.items():
            # Handle positional arguments
            if key.startswith("_"):
                self.template.append("{" + key + "}")

            # Handle optional ones
            else:
                arg = self.keys[key]
                if arg in self.template:
                    continue

                self.template.append(arg)

                # Ignore value as key is a boolean argument
                if isinstance(value, bool):
                    continue

                if not isinstance(value, list):
                    template = "{" + key + "}"
                    self.template.append(template)
                    continue

                for pos in range(len(value)):
                    template = "{" + key + "[" + str(pos) + "]}"
                    self.template.append(template)

        self._already_parsed = True

        return self.arguments

    @staticmethod
    def _key_to_arg(arg):
        if len(arg) > 1:
            return "--" + arg

        return "-" + arg

    def _parse_arguments(self, commandline):

        arguments = OrderedDict()
        keys = OrderedDict()
        key = None

        for item in commandline:
            # Handle keyworded arguments
            if item.startswith("-"):
                # Make sure we're not defining the same argument twice
                # If the argument is in the form of `--name=value`
                argument_parts = item.split('=')
                argument_name = argument_parts[0]
                key = argument_name.lstrip('-')

                if key in keys:
                    raise ValueError("Conflict: two arguments have the same name: {}"
                                     .format(key))

                arguments[key] = []
                keys[key] = argument_name

                if len(argument_parts) > 1:
                    arguments[key].append(argument_parts[-1])

            # If the argument did not start with `-` but we have an argument name
            # That means that this value belongs to that argument name list
            elif key is not None and item.strip(" "):
                arguments[key].append(item)

            # No argument name means we have not reached them yet, so we're still in the
            # Positional arguments part
            elif key is None:
                _pos_key = "_pos_{}".format(len(arguments))
                keys[_pos_key] = _pos_key
                arguments[_pos_key] = item

        for key, value in arguments.items():
            # Loop through the items and check if their value is a list
            # If it is, and the length is 0, that means it is a boolean args.
            # If its value is 1, it only has a single element and we unpack it.
            if isinstance(value, list):
                if not value:
                    value = True
                elif len(value) == 1:
                    value = value[0]

            value = self._parse_paths(value)
            arguments[key] = value

        return keys, arguments

    def _parse_paths(self, value):
        if isinstance(value, list):
            return [self._parse_paths(item) for item in value]

        if isinstance(value, str) and os.path.exists(value):
            return os.path.abspath(value)

        return value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.worker.transformer`."""
from collections import OrderedDict
import copy

import numpy
import pytest

from orion.algo.space import (Categorical, Dimension, Integer, Real, Space,)
from orion.core.worker.transformer import (build_required_space,
                                           Compose, Enumerate, Identity,
                                           OneHotEncode, Precision, Quantize, Reverse,
                                           TransformedDimension, TransformedSpace,)


class TestIdentity(object):
    """Test subclasses of `Identity` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Identity()
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Identity()
        assert t.domain_type is None
        assert t.target_type is None

        t = Identity('mpogias')
        assert t.domain_type == 'mpogias'
        assert t.target_type == 'mpogias'

    def test_transform(self):
        """Check if it transforms properly."""
        t = Identity()
        assert t.transform('yo') == 'yo'

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Identity()
        assert t.reverse('yo') == 'yo'

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Identity()
        assert t.infer_target_shape((5,)) == (5,)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Identity()
        assert t.repr_format('asfa') == 'asfa'


class TestReverse(object):
    """Test subclasses of `Reverse` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Reverse(Quantize())
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Reverse(Quantize())
        assert t.domain_type == 'integer'
        assert t.target_type == 'real'

    def test_transform(self):
        """Check if it transforms properly."""
        t = Reverse(Quantize())
        assert t.transform(9) == 9.
        assert t.transform(5) == 5.
        assert numpy.all(t.transform([9, 5]) == numpy.array([9., 5.], dtype=float))

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Reverse(Quantize())
        assert t.reverse(8.6) == 8
        assert t.reverse(5.3) == 5
        assert numpy.all(t.reverse([8.6, 5.3]) == numpy.array([8, 5], dtype=int))

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Reverse(Quantize())
        assert t.infer_target_shape((5,)) == (5,)

    def test_no_reverse_one_hot_encode(self):
        """Do NOT support real to categorical."""
        with pytest.raises(AssertionError):
            Reverse(OneHotEncode([1, 2, 3]))

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Reverse(Quantize())
        assert t.repr_format('asfa') == 'ReverseQuantize(asfa)'


class TestCompose(object):
    """Test subclasses of `Compose` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Compose([Enumerate([2, 'asfa', 'ipsi']), OneHotEncode(3)], 'categorical')
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Compose([])
        assert t.domain_type is None
        assert t.target_type is None

        t = Compose([], 'real')
        assert t.domain_type == 'real'
        assert t.target_type == 'real'

        t = Compose([Quantize()], 'real')
        assert t.domain_type == 'real'
        assert t.target_type == 'integer'

        t = Compose([Enumerate([2, 'asfa', 'ipsi']), OneHotEncode(3)], 'categorical')
        assert t.domain_type == 'categorical'
        assert t.target_type == 'real'

    def test_transform(self):
        """Check if it transforms properly."""
        t = Compose([Enumerate([2, 'asfa', 'ipsi']), OneHotEncode(3)], 'categorical')
        assert numpy.all(t.transform(2) == numpy.array((1., 0., 0.)))
        assert numpy.all(t.transform('asfa') == numpy.array((0., 1., 0.)))
        assert numpy.all(t.transform('ipsi') == numpy.array((0., 0., 1.)))
        with pytest.raises(KeyError):
            t.transform('aafdasfa')
        assert(numpy.all(t.transform([['ipsi', 'asfa'], [2, 'ipsi']]) ==
               numpy.array([[(0., 0., 1.), (0., 1., 0.)], [(1., 0., 0.), (0., 0., 1.)]])))

        t = Compose([Enumerate([2, 'asfa']), OneHotEncode(2)], 'categorical')
        assert t.transform(2) == 0.
        assert t.transform('asfa') == 1.
        with pytest.raises(KeyError):
            t.transform('ipsi')
        assert(numpy.all(t.transform([['asfa', 'asfa'], [2, 'asfa']]) ==
               numpy.array([[1., 1.], [0., 1.]])))

        # for the crazy enough
        t = Compose([Enumerate([2]), OneHotEncode(1)], 'categorical')
        assert t.transform(2) == 0.
        with pytest.raises(KeyError):
            t.transform('ipsi')
        assert numpy.all(t.transform([[2, 2], [2, 2]]) == [[0, 0], [0, 0]])

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Compose([Enumerate([2, 'asfa', 'ipsi']), OneHotEncode(3)], 'categorical')
        assert t.reverse((0.9, 0.8, 0.3)) == 2
        assert t.reverse((-0.3, 2., 0.)) == 'asfa'
        assert t.reverse((0., 0., 1.)) == 'ipsi'
        with pytest.raises(AssertionError):
            t.reverse((0., 0., 0., 1.))
        assert(numpy.all(t.reverse(numpy.array([[(0., 0., 1.), (0., 1., 0.)],
                                                [(1., 0., 0.), (0., 0., 1.)]])) ==
               numpy.array([['ipsi', 'asfa'], [2, 'ipsi']], dtype=numpy.object)))

        t = Compose([Enumerate([2, 'asfa']), OneHotEncode(2)], 'categorical')
        assert t.reverse(0.3) == 2
        assert t.reverse(2.) == 'asfa'
        assert numpy.all(t.reverse((0., 0., 0., 1.)) == numpy.array([2, 2, 2, 'asfa'],
                                                                    dtype=numpy.object))
        assert(numpy.all(t.reverse(numpy.array([[0.55, 3.], [-0.6, 1.]])) ==
               numpy.array([['asfa', 'asfa'], [2, 'asfa']], dtype=numpy.object)))

        # for the crazy enough
        t = Compose([Enumerate([2]), OneHotEncode(1)], 'categorical')
        assert t.reverse(0) == 2
        assert t.reverse(5.0) == 2
        assert t.reverse(0.2) == 2
        assert t.reverse(-0.2) == 2
        assert numpy.all(t.reverse([[0.5, 0], [1.0, 55]]) == numpy.array([[2, 2], [2, 2]],
                                                                         dtype=numpy.object))

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Compose([Enumerate([2, 'asfa', 'ipsi']), OneHotEncode(3)], 'categorical')
        assert t.infer_target_shape((2, 5)) == (2, 5, 3)

        t = Compose([Enumerate([2, 'asfa']), OneHotEncode(2)], 'categorical')
        assert t.infer_target_shape((2, 5)) == (2, 5)

        t = Compose([Enumerate([2]), OneHotEncode(1)], 'categorical')
        assert t.infer_target_shape((2, 5)) == (2, 5)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Compose([Enumerate([2, 'asfa', 'ipsi']), OneHotEncode(3)], 'categorical')
        assert t.repr_format('asfa') == 'OneHotEncode(Enumerate(asfa))'


class TestPrecision(object):
    """Test subclasses of `Precision` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Precision()
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Precision()
        assert t.domain_type == 'real'
        assert t.target_type == 'real'

    def test_transform(self):
        """Check if it transforms properly."""
        t = Precision(precision=4)
        assert t.transform(8.654321098) == 8.654
        assert t.transform(0.000123456789) == 0.0001235
        assert numpy.all(t.transform([8.654321098, 0.000123456789]) ==
                         numpy.array([8.654, 0.0001235], dtype=float))

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Precision()
        assert t.reverse(9.) == 9.
        assert t.reverse(5.) == 5.
        assert numpy.all(t.reverse([9., 5.]) == numpy.array([9., 5.], dtype=float))

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Precision()
        assert t.infer_target_shape((5,)) == (5,)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Precision()
        assert t.repr_format('asfa') == 'Precision(4, asfa)'


class TestQuantize(object):
    """Test subclasses of `Quantize` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Quantize()
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Quantize()
        assert t.domain_type == 'real'
        assert t.target_type == 'integer'

    def test_transform(self):
        """Check if it transforms properly."""
        t = Quantize()
        assert t.transform(8.6) == 8
        assert t.transform(5.3) == 5
        assert numpy.all(t.transform([8.6, 5.3]) == numpy.array([8, 5], dtype=int))

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Quantize()
        assert t.reverse(9) == 9.
        assert t.reverse(5) == 5.
        assert numpy.all(t.reverse([9, 5]) == numpy.array([9., 5.], dtype=float))

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Quantize()
        assert t.infer_target_shape((5,)) == (5,)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Quantize()
        assert t.repr_format('asfa') == 'Quantize(asfa)'


class TestEnumerate(object):
    """Test subclasses of `Enumerate` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = Enumerate([2, 'asfa', 'ipsi'])
        # Copy won't fail if vectorized function is not called at least once.
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = Enumerate([2, 'asfa', 'ipsi'])
        assert t.domain_type == 'categorical'
        assert t.target_type == 'integer'

    def test_transform(self):
        """Check if it transforms properly."""
        t = Enumerate([2, 'asfa', 'ipsi'])
        assert t.transform(2) == 0
        assert t.transform('asfa') == 1
        assert t.transform('ipsi') == 2
        with pytest.raises(KeyError):
            t.transform('aafdasfa')
        assert numpy.all(t.transform([['ipsi', 'asfa'], [2, 'ipsi']]) == [[2, 1], [0, 2]])

        # for the crazy enough
        t = Enumerate([2])
        assert t.transform(2) == 0
        with pytest.raises(KeyError):
            t.transform('aafdasfa')
        assert numpy.all(t.transform([[2, 2], [2, 2]]) == [[0, 0], [0, 0]])

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = Enumerate([2, 'asfa', 'ipsi'])
        assert t.reverse(0) == 2
        assert t.reverse(1) == 'asfa'
        assert t.reverse(2) == 'ipsi'
        with pytest.raises(IndexError):
            t.reverse(3)
        assert numpy.all(t.reverse([[2, 1], [0, 2]]) == numpy.array([['ipsi', 'asfa'], [2, 'ipsi']],
                                                                    dtype=numpy.object))

        # for the crazy enough
        t = Enumerate([2])
        assert t.reverse(0) == 2
        with pytest.raises(IndexError):
            t.reverse(1)
        assert numpy.all(t.reverse([[0, 0], [0, 0]]) == numpy.array([[2, 2], [2, 2]],
                                                                    dtype=numpy.object))

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = Enumerate([2, 'asfa', 'ipsi'])
        assert t.infer_target_shape((5,)) == (5,)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = Enumerate([2, 'asfa', 'ipsi'])
        assert t.repr_format('asfa') == 'Enumerate(asfa)'


class TestOneHotEncode(object):
    """Test subclasses of `OneHotEncode` transformation."""

    def test_deepcopy(self):
        """Verify that the transformation object can be copied"""
        t = OneHotEncode(3)
        t.transform([2])
        copy.deepcopy(t)

    def test_domain_and_target_type(self):
        """Check if attribute-like `domain_type` and `target_type` do
        what's expected.
        """
        t = OneHotEncode(3)
        assert t.domain_type == 'integer'
        assert t.target_type == 'real'

    def test_transform(self):
        """Check if it transforms properly."""
        t = OneHotEncode(3)
        assert numpy.all(t.transform(0) == numpy.array((1., 0., 0.)))
        assert numpy.all(t.transform(1) == numpy.array((0., 1., 0.)))
        assert numpy.all(t.transform(2) == numpy.array((0., 0., 1.)))
        with pytest.raises(AssertionError):
            t.transform(4)
        with pytest.raises(AssertionError):
            t.transform(-1)
        with pytest.raises(AssertionError):
            t.transform(2.2)
        assert(numpy.all(t.transform([[2, 1], [0, 2]]) ==
               numpy.array([[(0., 0., 1.), (0., 1., 0.)], [(1., 0., 0.), (0., 0., 1.)]])))

        t = OneHotEncode(2)
        assert t.transform(0) == 0.
        assert t.transform(1) == 1.
        with pytest.raises(TypeError):
            t.transform('ipsi')
        assert(numpy.all(t.transform([[1, 1], [0, 1]]) ==
               numpy.array([[1., 1.], [0., 1.]])))

        # for the crazy enough
        t = OneHotEncode(1)
        assert t.transform(0) == 0.
        with pytest.raises(TypeError):
            t.transform('ipsi')
        assert numpy.all(t.transform([[0, 0], [0, 0]]) == [[0., 0.], [0., 0.]])

    def test_reverse(self):
        """Check if it reverses `transform` properly, if possible."""
        t = OneHotEncode(3)
        assert t.reverse((0.9, 0.8, 0.3)) == 0
        assert t.reverse((-0.3, 2., 0.)) == 1
        assert t.reverse((0., 0., 1.)) == 2
        with pytest.raises(AssertionError):
            t.reverse((0., 0., 0., 1.))
        assert(numpy.all(t.reverse(numpy.array([[[0., 0., 1.], [0., 1., 0.]],
                                                [[1., 0., 0.], [0., 0., 1.]]])) ==
               numpy.array([[2, 1], [0, 2]], dtype=int)))

        t = OneHotEncode(2)
        assert t.reverse(0.3) == 0
        assert t.reverse(2.) == 1
        assert numpy.all(t.reverse((0., 0., 0., 1.)) == numpy.array([0, 0, 0, 1],
                                                                    dtype=int))
        assert(numpy.all(t.reverse(numpy.array([[0.55, 3.], [-0.6, 1.]])) ==
               numpy.array([[1, 1], [0, 1]], dtype=int)))

        # for the crazy enough
        t = OneHotEncode(1)
        assert t.reverse(0) == 0
        assert t.reverse(5.0) == 0
        assert t.reverse(0.2) == 0
        assert t.reverse(-0.2) == 0
        assert numpy.all(t.reverse([[0.5, 0], [1.0, 55]]) == numpy.array([[0, 0], [0, 0]],
                                                                         dtype=int))

    def test_interval(self):
        """Test that the onehot interval has the proper dimensions"""
        t = OneHotEncode(3)
        low, high = t.interval()
        assert (low == numpy.zeros(3)).all()
        assert (high == numpy.ones(3)).all()

        t = OneHotEncode(2)
        low, high = t.interval()
        assert (low == numpy.zeros(1)).all()
        assert (high == numpy.ones(1)).all()

    def test_infer_target_shape(self):
        """Check if it infers the shape of a transformed `Dimension`."""
        t = OneHotEncode(3)
        assert t.infer_target_shape((2, 5)) == (2, 5, 3)

        t = OneHotEncode(2)
        assert t.infer_target_shape((2, 5)) == (2, 5)

        t = OneHotEncode(1)
        assert t.infer_target_shape((2, 5)) == (2, 5)

    def test_repr_format(self):
        """Check representation of a transformed dimension."""
        t = OneHotEncode(3)
        assert t.repr_format('asfa') == 'OneHotEncode(asfa)'


@pytest.fixture(scope='module')
def dim():
    """Create an example of `Dimension`."""
    dim = Real('yolo', 'norm', 0.9, shape=(3, 2))
    return dim


@pytest.fixture(scope='module')
def tdim(dim):
    """Create an example of `TransformedDimension`."""
    transformers = [Quantize()]
    tdim = TransformedDimension(Compose(transformers, dim.type),
                                dim)
    return tdim


@pytest.fixture(scope='module')
def dim2():
    """Create a second example of `Dimension`."""
    probs = (0.1, 0.2, 0.3, 0.4)
    categories = ('asdfa', '2', '3', '4')
    categories = OrderedDict(zip(categories, probs))
    dim2 = Categorical('yolo2', categories)
    return dim2


@pytest.fixture(scope='module')
def tdim2(dim2):
    """Create a second example of `TransformedDimension`."""
    transformers = [Enumerate(dim2.categories), OneHotEncode(len(dim2.categories))]
    tdim2 = TransformedDimension(Compose(transformers, dim2.type),
                                 dim2)
    return tdim2


class TestTransformedDimension(object):
    """Check functionality of class `TransformedDimension`."""

    def test_transform(self, tdim):
        """Check method `transform`."""
        assert tdim.transform(8.6) == 8
        assert tdim.transform(5.3) == 5
        assert numpy.all(tdim.transform([8.6, 5.3]) == numpy.array([8, 5], dtype=int))

    def test_reverse(self, tdim):
        """Check method `reverse`."""
        assert tdim.reverse(9) == 9.
        assert tdim.reverse(5) == 5.
        assert numpy.all(tdim.reverse([9, 5]) == numpy.array([9., 5.], dtype=float))

    def test_mimics_Dimension(self, tdim):
        """Mimic `Dimension`.
        Set of `Dimension`'s methods are subset of `TransformedDimension`.
        """
        transformed_dimension_keys = set(TransformedDimension.__dict__.keys())
        # For some reason running all tests have the side-effect of adding an attribute
        # __slotnames__ to TransformedDimension. This attribute is not present when running
        # tests found in test_transformer.py only.
        transformed_dimension_keys.discard('__slotnames__')
        assert ((transformed_dimension_keys ^ set(Dimension.__dict__.keys())) ==
                set(['transform', 'reverse']))

    def test_sample(self, tdim, seed):
        """Check method `sample`."""
        assert numpy.all(tdim.sample(seed=seed) == numpy.array([[1, 0], [3, 0], [1, 2]]))
        samples = tdim.sample(2, seed=seed)
        assert len(samples) == 2
        assert numpy.all(samples[0] == numpy.array([[-1, 0], [1, 0], [-1, 0]]))
        assert numpy.all(samples[1] == numpy.array([[0, 1], [-1, 0], [2, 2]]))

    def test_interval(self, tdim):
        """Check method `interval`."""
        tmp1 = tdim.original_dimension._low
        tmp2 = tdim.original_dimension._high
        tdim.original_dimension._low = -0.6
        tdim.original_dimension._high = 1.2

        assert tdim.interval() == (-1, 1)

        tdim.original_dimension._low = tmp1
        tdim.original_dimension._high = tmp2

    def test_interval_from_categorical(self, tdim2):
        """Check how we should treat interval when original dimension is categorical."""
        low, high = tdim2.interval()
        assert (low == numpy.zeros(4)).all()
        assert (high == numpy.ones(4)).all()

    def test_contains(self, tdim):
        """Check method `__contains__`."""
        assert [[1, 1], [3, 1], [1, 2]] in tdim

        tmp1 = tdim.original_dimension._low
        tmp2 = tdim.original_dimension._high
        tdim.original_dimension._low = -0.6
        tdim.original_dimension._high = 1.2

        assert [[1, 1], [3, 1], [1, 2]] not in tdim

        tdim.original_dimension._low = tmp1
        tdim.original_dimension._high = tmp2

    def test_contains_from_categorical(self, tdim2):
        """Check method `__contains__` when original is categorical."""
        assert (0, 0, 0, 1) in tdim2
        assert (0, 2, 0, 1) in tdim2
        assert (0, 2, 0) not in tdim2

    def test_eq(self, tdim, tdim2):
        """Return True if other is the same transformed dimension as self"""
        assert tdim != tdim2
        assert tdim == copy.deepcopy(tdim)

    def test_hash(self, tdim, tdim2):
        """Test that hash is consistent for identical and different transformed dimensions"""
        assert hash(tdim) != hash(tdim2)
        assert hash(tdim) == hash(copy.deepcopy(tdim))

    def test_get_hashable_members(self, tdim, tdim2):
        """Test that hashable members of the transformed dimensions are the aggregation of
        transformer's and original dimension's hashable members.
        """
        assert (tdim._get_hashable_members() ==
                ('Compose', 'Quantize', 'real', 'integer', 'Identity', 'real', 'real',
                 'yolo', (3, 2), 'real', (0.9,), (), None, 'norm'))
        assert (tdim2._get_hashable_members() ==
                ('Compose', 'OneHotEncode', 'integer', 'real', 4, 'Compose', 'Enumerate',
                 'categorical', 'integer', 'Identity', 'categorical', 'categorical', 'yolo2', (),
                 'categorical', (), (), None, 'Distribution'))

    def test_validate(self, tdim, tdim2):
        """Validate original_dimension"""
        # It pass
        tdim.validate()
        tdim2.validate()

        # We break it
        tdim.original_dimension._kwargs['size'] = (2, )
        tdim2.original_dimension._default_value = 'bad-default'

        # It does not pass
        with pytest.raises(ValueError) as exc:
            tdim.validate()
        assert "Use 'shape' keyword only instead of 'size'." in str(exc.value)

        with pytest.raises(ValueError) as exc:
            tdim2.validate()
        assert "bad-default is not a valid value for this Dimension." in str(exc.value)

        tdim.original_dimension._kwargs.pop('size')
        tdim2.original_dimension._default_value = Dimension.NO_DEFAULT_VALUE

    def test_get_prior_string(self, tdim, tdim2):
        """Apply the transformation on top of the prior string of original dimension."""
        assert tdim.get_prior_string() == "Quantize(norm(0.9, shape=(3, 2)))"
        assert tdim2.get_prior_string() == "OneHotEncode(Enumerate(choices({'asdfa': 0.10, '2': 0.20, '3': 0.30, '4': 0.40})))"  # noqa

    def test_get_string(self, tdim, tdim2):
        """Apply the transformation only on top of the prior string of original dimension."""
        assert tdim.get_string() == "yolo~Quantize(norm(0.9, shape=(3, 2)))"
        assert tdim2.get_string() == "yolo2~OneHotEncode(Enumerate(choices({'asdfa': 0.10, '2': 0.20, '3': 0.30, '4': 0.40})))"  # noqa

    def test_repr(self, tdim):
        """Check method `__repr__`."""
        assert str(tdim) == "Quantize(Real(name=yolo, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None))"  # noqa

    def test_name_property(self, tdim):
        """Check property `name`."""
        assert tdim.name == 'yolo'

    def test_type_property(self, tdim, tdim2):
        """Check property `type`."""
        assert tdim.type == 'integer'
        assert tdim2.type == 'real'

    def test_prior_name_property(self, tdim, tdim2):
        """Check property `prior_name`."""
        assert tdim.prior_name == 'norm'
        assert tdim2.prior_name == 'choices'

    def test_shape_property(self, tdim, tdim2):
        """Check property `shape`."""
        assert tdim.original_dimension.shape == (3, 2)
        assert tdim.shape == (3, 2)
        assert tdim2.original_dimension.shape == ()
        assert tdim2.shape == (4,)

    def test_default_value_property(self, tdim, tdim2):
        """Check property `default_value`."""
        assert tdim.default_value is None
        tdim2.original_dimension._default_value = '3'
        assert numpy.all(tdim2.default_value == (0., 0., 1., 0.))
        tdim2.original_dimension._default_value = None

    def test_cast(self, tdim, tdim2):
        """Check casting through transformers"""
        assert tdim.cast(['10.1']) == [10.0]
        assert numpy.all(tdim2.cast(['asdfa']) == numpy.array([[1, 0, 0, 0]]))
        assert numpy.all(tdim2.cast(['3']) == numpy.array([[0, 0, 1, 0]]))


@pytest.fixture(scope='module')
def space(dim, dim2):
    """Create an example `Space`."""
    space = Space()
    space.register(dim)
    space.register(dim2)
    return space


@pytest.fixture(scope='module')
def tspace(tdim, tdim2):
    """Create an example `TransformedSpace`."""
    tspace = TransformedSpace()
    tspace.register(tdim)
    tspace.register(tdim2)
    return tspace


class TestTransformedSpace(object):
    """Check functionality of class `TransformedSpace`."""

    def test_extends_space(self, tspace):
        """Check that `TransformedSpace` is actually a `Space`."""
        assert isinstance(tspace, Space)

    def test_transform(self, space, tspace, seed):
        """Check method `transform`."""
        yo = space.sample(seed=seed)[0]
        tyo = tspace.transform(yo)
        assert tyo in tspace

    def test_reverse(self, space, tspace, seed):
        """Check method `reverse`."""
        tyo = tspace.sample(seed=seed)[0]
        yo = tspace.reverse(tyo)
        assert yo in space


@pytest.fixture(scope='module')
def space_each_type(dim, dim2):
    """Create an example `Space`."""
    space = Space()
    space.register(dim)
    space.register(dim2)
    space.register(Integer('yolo3', 'randint', 3, 10))
    return space


class TestRequiredSpaceBuilder(object):
    """Check functionality of builder function `build_required_space`."""

    def test_not_supported_requirement(self, space_each_type):
        """Require something which is not supported."""
        with pytest.raises(TypeError) as exc:
            build_required_space('fasdfasf', space_each_type)
        assert 'Unsupported' in str(exc.value)

    def test_no_requirement(self, space_each_type):
        """Check what is built using 'None' requirement."""
        tspace = build_required_space(None, space_each_type)
        assert len(tspace) == 3
        assert tspace[0].type == 'real'
        assert tspace[1].type == 'categorical'
        assert tspace[2].type == 'integer'
        assert (str(tspace) ==
                "Space([Precision(4, Real(name=yolo, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None)),\n"  # noqa
                "       Categorical(name=yolo2, prior={asdfa: 0.10, 2: 0.20, 3: 0.30, 4: 0.40}, shape=(), default value=None),\n"  # noqa
                "       Integer(name=yolo3, prior={randint: (3, 10), {}}, shape=(), default value=None)])")  # noqa

        tspace = build_required_space([], space_each_type)
        assert len(tspace) == 3
        assert tspace[0].type == 'real'
        assert tspace[1].type == 'categorical'
        assert tspace[2].type == 'integer'
        assert (str(tspace) ==
                "Space([Precision(4, Real(name=yolo, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None)),\n"  # noqa
                "       Categorical(name=yolo2, prior={asdfa: 0.10, 2: 0.20, 3: 0.30, 4: 0.40}, shape=(), default value=None),\n"  # noqa
                "       Integer(name=yolo3, prior={randint: (3, 10), {}}, shape=(), default value=None)])")  # noqa

    def test_integer_requirement(self, space_each_type):
        """Check what is built using 'integer' requirement."""
        tspace = build_required_space('integer', space_each_type)
        assert len(tspace) == 3
        assert tspace[0].type == 'integer'
        assert tspace[1].type == 'integer'
        assert tspace[2].type == 'integer'
        assert(str(tspace) ==
               "Space([Quantize(Real(name=yolo, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None)),\n"  # noqa
               "       Enumerate(Categorical(name=yolo2, prior={asdfa: 0.10, 2: 0.20, 3: 0.30, 4: 0.40}, shape=(), default value=None)),\n"  # noqa
               "       Integer(name=yolo3, prior={randint: (3, 10), {}}, shape=(), default value=None)])")  # noqa

    def test_real_requirement(self, space_each_type):
        """Check what is built using 'real' requirement."""
        tspace = build_required_space('real', space_each_type)
        assert len(tspace) == 3
        assert tspace[0].type == 'real'
        assert tspace[1].type == 'real'
        assert tspace[2].type == 'real'
        assert(str(tspace) ==
               "Space([Precision(4, Real(name=yolo, prior={norm: (0.9,), {}}, shape=(3, 2), default value=None)),\n"  # noqa
               "       OneHotEncode(Enumerate(Categorical(name=yolo2, prior={asdfa: 0.10, 2: 0.20, 3: 0.30, 4: 0.40}, shape=(), default value=None))),\n"  # noqa
               "       ReverseQuantize(Integer(name=yolo3, prior={randint: (3, 10), {}}, shape=(), default value=None))])")  # noqa

    def test_capacity(self, space_each_type):
        """Check transformer space capacity"""
        tspace = build_required_space('real', space_each_type)
        assert tspace.cardinality == numpy.inf

        space = Space()
        probs = (0.1, 0.2, 0.3, 0.4)
        categories = ('asdfa', 2, 3, 4)
        dim = Categorical('yolo', OrderedDict(zip(categories, probs)), shape=2)
        space.register(dim)
        dim = Integer('yolo2', 'uniform', -3, 6)
        space.register(dim)
        tspace = build_required_space('integer', space)
        assert tspace.cardinality == (4 * 2) * 6

        dim = Integer('yolo3', 'uniform', -3, 6, shape=(2, 1))
        space.register(dim)
        tspace = build_required_space('integer', space)
        assert tspace.cardinality == (4 * 2) * 6 * 6 * (2 * 1)


def test_quantization_does_not_violate_bounds():
    """Regress on bug that converts valid float in tdim to non valid excl. upper bound."""
    dim = Integer('yo', 'uniform', 3, 7)
    transformers = [Reverse(Quantize())]
    tdim = TransformedDimension(Compose(transformers, dim.type), dim)
    assert 11 not in dim
    assert 10 in dim
    # but be careful, because upper bound is inclusive
    assert 11.5 not in tdim
    assert 10.6 in tdim
    assert tdim.reverse(9.6) in dim
    # solution is to quantize with 'floor' instead of 'round'
    assert tdim.reverse(9.6) == 9



# -*- coding: utf-8 -*-
# pylint: skip-file
"""
:mod:`orion.core.worker.trial` -- Container class for `Trial` entity
====================================================================

.. module:: trial
   :platform: Unix
   :synopsis: Describe a particular training run, parameters and results

"""
import hashlib
import logging

from orion.core.utils.flatten import unflatten


log = logging.getLogger(__name__)


def validate_status(status):
    """
    Verify if given status is valid. Can be one of ``new``, ``reserved``, ``suspended``,
    ``completed``, ``interrupted``, or ``broken``.
    """
    if status is not None and status not in Trial.allowed_stati:
        raise ValueError("Given status `{0}` not one of: {1}".format(
            status, Trial.allowed_stati))


class Trial:
    """Represents an entry in database/trials collection.

    Attributes
    ----------
    experiment : str
       Unique identifier for the experiment that produced this trial.
       Same as an `Experiment._id`.
    id_override: str
        Trial id returned by the database. It should be unique for a given
        set of parameters
    heartbeat : datetime.datetime
        Last time trial was identified as being alive.
    status : str
       Indicates how this trial is currently being used. Can take the following
       values:

       * 'new' : Denotes a fresh set of parameters suggested by an algorithm,
          not yet tried out.
       * 'reserved' : Indicates that this trial is currently being evaluated by
          a worker process, it was a 'new' trial that got selected.
       * 'suspended' : Means that an algorithm decided to stop the evaluation of
          a 'reserved' trial prematurely.
       * 'completed' : is the status of a previously 'reserved' trial that
          successfully got evaluated. `Trial.results` must contain the evaluation.
       * 'interrupted' : Indicates trials that are stopped from being evaluated
          by external *actors* (e.g. cluster timeout, KeyboardInterrupt, killing
          of the worker process).
       * 'broken' : Indicates a trial that was not successfully evaluated for not
          expected reason.
    worker : str
       Corresponds to worker's unique id that handled this trial.
    submit_time : `datetime.datetime`
       When was this trial suggested?
    start_time : `datetime.datetime`
       When was this trial first reserved?
    end_time : `datetime.datetime`
       When was this trial evaluated successfully?
    results : list of `Trial.Result`
       List of evaluated metrics for this particular set of params. One and only
       one of them is necessarily an *objective* function value. The other are
       *constraints*, the value of an expression desired to be larger/equal to 0.
    params : dict of params
       Dict of suggested values for the `Experiment` parameter space.
       Consists a sample to be evaluated.

    """

    @classmethod
    def build(cls, trial_entries):
        """Builder method for a list of trials.

        :param trial_entries: List of trial representation in dictionary form,
           as expected to be saved in a database.

        :returns: a list of corresponding `Trial` objects.
        """
        trials = []
        for entry in trial_entries:
            trials.append(cls(**entry))
        return trials

    class Value:
        """Container for a value object.

        Attributes
        ----------
        name : str
           A possible named for the quality that this is quantifying.
        type : str
           An identifier with semantic importance for **Oríon**. See
           `Param.type` and `Result.type`.
        value : str or numerical
           value suggested for this dimension of the parameter space.

        """

        __slots__ = ('name', '_type', 'value')
        allowed_types = ()

        def __init__(self, **kwargs):
            """See attributes of `Value` for possible argument for `kwargs`."""
            for attrname in self.__slots__:
                setattr(self, attrname, None)
            for attrname, value in kwargs.items():
                setattr(self, attrname, value)

            self._ensure_no_ndarray()

        def _ensure_no_ndarray(self):
            """Make sure the current value is not a `numpy.ndarray`."""
            if hasattr(self, 'value') and hasattr(self.value, 'tolist'):
                self.value = self.value.tolist()

        def to_dict(self):
            """Needed to be able to convert `Value` to `dict` form."""
            ret = dict(
                name=self.name,
                type=self.type,
                value=self.value
                )
            return ret

        def __eq__(self, other):
            """Test equality based on self.to_dict()"""
            return self.name == other.name and self.type == other.type and self.value == other.value

        def __str__(self):
            """Represent partially with a string."""
            ret = "{0}(name={1}, type={2}, value={3})".format(
                type(self).__name__, repr(self.name),
                repr(self.type), repr(self.value))
            return ret

        __repr__ = __str__

        @property
        def type(self):
            """For meaning of property type, see `Value.type`."""
            return self._type

        @type.setter
        def type(self, type_):
            if type_ is not None and type_ not in self.allowed_types:
                raise ValueError("Given type, {0}, not one of: {1}".format(
                    type_, self.allowed_types))
            self._type = type_

    class Result(Value):
        """Types for a `Result` can be either an evaluation of an 'objective'
        function or of an 'constraint' expression.
        """

        __slots__ = ()
        allowed_types = ('objective', 'constraint', 'gradient', 'statistic', 'lie')

    class Param(Value):
        """Types for a `Param` can be either an integer (discrete value),
        floating precision numerical or a categorical expression (e.g. a string).
        """

        __slots__ = ()
        allowed_types = ('integer', 'real', 'categorical', 'fidelity')

    __slots__ = ('experiment', '_id', '_status', 'worker', '_working_dir', 'heartbeat',
                 'submit_time', 'start_time', 'end_time', '_results', '_params', 'parents',
                 'id_override')
    allowed_stati = ('new', 'reserved', 'suspended', 'completed', 'interrupted', 'broken')

    def __init__(self, **kwargs):
        """See attributes of `Trial` for meaning and possible arguments for `kwargs`."""
        for attrname in self.__slots__:
            if attrname in ('_results', '_params', 'parents'):
                setattr(self, attrname, list())
            else:
                setattr(self, attrname, None)

        self.status = 'new'

        # Store the id as an override to support different backends
        self.id_override = kwargs.pop('_id', None)

        for attrname, value in kwargs.items():
            if attrname == 'results':
                attr = getattr(self, attrname)
                for item in value:
                    attr.append(self.Result(**item))
            elif attrname == 'params':
                for item in value:
                    self._params.append(self.Param(**item))
            else:
                setattr(self, attrname, value)

    def to_dict(self):
        """Needed to be able to convert `Trial` to `dict` form."""
        trial_dictionary = dict()

        for attrname in self.__slots__:
            if attrname == "_working_dir":
                continue

            attrname = attrname.lstrip("_")
            trial_dictionary[attrname] = getattr(self, attrname)

        # Overwrite "results" and "params" with list of dictionaries rather
        # than list of Value objects
        trial_dictionary['results'] = list(map(lambda x: x.to_dict(), self.results))
        trial_dictionary['params'] = list(map(lambda x: x.to_dict(), self._params))

        trial_dictionary['_id'] = trial_dictionary.pop('id')

        return trial_dictionary

    def __str__(self):
        """Represent partially with a string."""
        return "Trial(experiment={0}, status={1}, params={2})".format(
            repr(self.experiment), repr(self._status), self.format_params(self._params))

    __repr__ = __str__

    @property
    def params(self):
        """Parameters of the trial"""
        return unflatten({param.name: param.value for param in self._params})

    @property
    def results(self):
        """List of results of the trial"""
        return self._results

    @results.setter
    def results(self, results):
        """Verify results before setting the property"""
        objective = self._fetch_one_result_of_type('objective', results)

        if objective is None:
            raise ValueError('No objective found in results: {}'.format(results))
        if not isinstance(objective.value, (float, int)):
            raise ValueError(
                'Results must contain a type `objective` with type float/int: {}'.format(objective))

        self._results = results

    @property
    def working_dir(self):
        """Return the current working directory of the trial."""
        return self._working_dir

    @working_dir.setter
    def working_dir(self, value):
        """Change the current working directory of the trial."""
        self._working_dir = value

    @property
    def status(self):
        """For meaning of property type, see `Trial.status`."""
        return self._status

    @status.setter
    def status(self, status):
        validate_status(status)
        self._status = status

    @property
    def id(self):
        """Return hash_name which is also the database key `_id`."""
        if self.id_override is None:
            return self.__hash__()
        return self.id_override

    @property
    def objective(self):
        """Return this trial's objective value if it is evaluated, else None.

        :rtype: `Trial.Result`
        """
        return self._fetch_one_result_of_type('objective')

    @property
    def lie(self):
        """Return this trial's fake objective value if it was set, else None.

        :rtype: `Trial.Result`
        """
        return self._fetch_one_result_of_type('lie')

    @property
    def gradient(self):
        """Return this trial's gradient value if it is evaluated, else None.

        :rtype: `Trial.Result`
        """
        return self._fetch_one_result_of_type('gradient')

    @property
    def constraints(self):
        """
        Return this trial's constraints

        Returns
        -------
        A list of ``Trial.Result`` of type 'constraint'
        """
        return self._fetch_results('constraint', self.results)

    @property
    def statistics(self):
        """
        Return this trial's statistics

        Returns
        -------
        A list of ``Trial.Result`` de type 'statistic'
        """
        return self._fetch_results('statistic', self.results)

    @property
    def hash_name(self):
        """Generate a unique name with an md5sum hash for this `Trial`.

        .. note:: Two trials that have the same `params` must have the same `hash_name`.
        """
        return self.compute_trial_hash(self, ignore_fidelity=False)

    @property
    def hash_params(self):
        """Generate a unique param md5sum hash for this `Trial`.

        .. note:: The params contributing to the hash do not include the fidelity.
        """
        return self.compute_trial_hash(self, ignore_fidelity=True, ignore_lie=True)

    def __hash__(self):
        """Return the hashname for this trial"""
        return self.hash_name

    @property
    def full_name(self):
        """Generate a unique name using the full definition of parameters."""
        if not self._params or not self.experiment:
            raise ValueError("Cannot distinguish this trial, as 'params' or 'experiment' "
                             "have not been set.")
        return self.format_values(self._params, sep='-').replace('/', '.')

    def _fetch_results(self, type, results):
        """Fetch results for the given type"""
        return [result for result in results if result.type == type]

    def _fetch_one_result_of_type(self, result_type, results=None):
        if results is None:
            results = self.results

        value = self._fetch_results(result_type, results)

        if not value:
            return None

        if len(value) > 1:
            log.warning("Found multiple results of '%s' type:\n%s",
                        result_type, value)
            log.warning("Multi-objective optimization is not currently supported.\n"
                        "Optimizing according to the first one only: %s", value[0])

        return value[0]

    def _repr_values(self, values, sep=','):
        """Represent with a string the given values."""
        return Trial.format_values(values, sep)

    def params_repr(self, sep=',', ignore_fidelity=False):
        """Represent with a string the parameters contained in this `Trial` object."""
        return Trial.format_params(self._params, sep)

    @staticmethod
    def format_values(values, sep=','):
        """Represent with a string the given values."""
        return sep.join(map(lambda value: "{0.name}:{0.value}".format(value), values))

    @staticmethod
    def format_params(params, sep=',', ignore_fidelity=False):
        """Represent with a string the parameters contained in this `Trial` object."""
        if ignore_fidelity:
            params = [x for x in params if x.type != 'fidelity']
        else:
            params = params
        return Trial.format_values(params, sep)

    @staticmethod
    def compute_trial_hash(trial, ignore_fidelity=False, ignore_experiment=False,
                           ignore_lie=False):
        """Generate a unique param md5sum hash for a given `Trial`"""
        if not trial._params and not trial.experiment:
            raise ValueError("Cannot distinguish this trial, as 'params' or 'experiment' "
                             "have not been set.")

        params = Trial.format_params(trial._params, ignore_fidelity=ignore_fidelity)

        experiment_repr = ""
        if not ignore_experiment:
            experiment_repr = str(trial.experiment)

        lie_repr = ""
        if not ignore_lie and trial.lie:
            lie_repr = Trial.format_values([trial.lie])

        return hashlib.md5((params + experiment_repr + lie_repr).encode('utf-8')).hexdigest()



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.evc.tree` -- Tree data structure for the experiment version control system
===========================================================================================

.. module:: evc
   :platform: Unix
   :synopsis: Tree data structure for the experiment version control system

Experiment version control requires building trees of the experiments so
that we can fetch trials from one experiment to another or navigate from
one experiment to another to visualise different statistics.

TreeNode and tree iterators support the tree data structure of the experiment version control.
TreeNode is a generic class which can carry arbitrary python objects. It comes with basic methods to
set parent and children. A method `map` allows to apply functions recursively on the tree in a
generic manner.

"""


# pylint: disable=too-few-public-methods
class PreOrderTraversal(object):
    """Iterate on a tree in a pre-order traversal fashion

    Attributes
    ----------
    stack: list of `orion.core.evc.tree.TreeNode`
        Nodes logged during iteration

    """

    __slots__ = ('stack', )

    def __init__(self, tree_node):
        """Initialize the stack for iteration"""
        self.stack = [tree_node]

    def __iter__(self):
        """Get the iterator"""
        return self

    def __next__(self):
        """Get the next node in pre-order traversal"""
        try:
            node = self.stack.pop()
        except IndexError:
            raise StopIteration

        self.stack += node.children[::-1]

        return node


# pylint: disable=too-few-public-methods
class DepthFirstTraversal(object):
    """Iterate on a tree in a pre-order traversal fashion

    Attributes
    ----------
    stack: list of `orion.core.evc.tree.TreeNode`
        Nodes logged during iteration
    seen: set of `orion.core.evc.tree.TreeNode`
        Nodes which have been returned during iteration

    """

    __slots__ = ('stack', 'seen')

    def __init__(self, tree_node):
        """Initialize the stack and set of seen nodes for iteration"""
        self.stack = [tree_node]
        self.seen = set()

    def _compute_potential(self):
        """Filter out seen nodes from the stack"""
        if not self.stack:
            return []

        return list(filter(lambda n: n not in self.seen, self.stack[-1].children))

    def __iter__(self):
        """Get the iterator"""
        return self

    def __next__(self):
        """Get the next node in depth-first traversal"""
        potential = self._compute_potential()
        while self.stack and potential:
            self.stack += potential[::-1]
            potential = self._compute_potential()

        try:
            node = self.stack.pop()
        except IndexError:
            raise StopIteration

        self.seen.add(node)

        return node


class TreeNode(object):
    r"""Tree node data structure

    Nodes have an attribute item to carry arbitrary information. A node may only have one parent and
    can have as many children as desired.

    Parents can be set at initialiation or via `node.set_parent`. Setting a parent automatically set
    the current node as the child of the parent.

    Children can be set at initialiation or via `node.add_children`. Setting children automatically
    set their parent as the current node.

    Tree of nodes are iterable, by default with preorder traversal.

    .. seealso::
        `orion.core.evc.tree.PreOrderTraversal`
        `orion.core.evc.tree.DepthFirstTraversal`

    Attributes
    ----------
    item: object
        Can be anything
    parent: None or instance of `orion.core.evc.tree.TreeNode`
        The parent of the current node, None if the current node is the root.
    children: None or list of instances of `orion.core.evc.tree.TreeNode`
        The children of the curent node.
    root: instance of `orion.core.evc.tree.TreeNode`
        The top node of the current tree. The root node returns itself.

    Examples
    --------
    .. code-block:: python
        :linenos:

        a = TreeNode("a")
        b = TreeNode("b", a)
        c = TreeNode("c", a)
        d = TreeNode("d", a)
        e = TreeNode("e", a)

        f = TreeNode("f", b)
        g = TreeNode("g", b)

        h = TreeNode("h", e)

        # Gives this tree
        # a
        # |   \  \   \
        # b    c  d   e
        # | \         |
        # f  g        h

        g.set_parent(e)

        # Gives this tree
        # a
        # |  \  \   \
        # b   c  d   e
        # |          | \
        # f          h g

        c.add_children(h, g)

        # Gives this tree
        # a
        # |  \    \   \
        # b   c    d   e
        # |   | \
        # f   h  g

        a.drop_children(c)

        # Gives this tree
        # a
        # |  \   \
        # b   d   e
        # |
        # f

    """

    __slots__ = ('_item', '_parent', '_children')

    def __init__(self, item, parent=None, children=tuple()):
        """Initialize node with item, parent and children

        .. seealso::
            :class:`orion.core.evc.tree.TreeNode` for information about the attributes
        """
        self._item = item
        self._parent = None
        self._children = []

        if parent is not None:
            self.set_parent(parent)
        if children:
            self.add_children(*children)

    @property
    def item(self):
        """Get item of the node which may contain arbitrary objects"""
        return self._item

    @item.setter
    def item(self, new_item):
        """Set item of the node with arbitrary objects"""
        self._item = new_item

    @property
    def parent(self):
        """Get parent of the node, None if no parent"""
        return self._parent

    def drop_parent(self):
        """Drop the parent of the node, do nothing if no parent

        Note that the node will be removed from the children of the parent as well
        """
        if self.parent is not None:
            self._parent.drop_children(self)

    def set_parent(self, node):
        """Set the parent of the node

        Note that setting a new parent will have the effect of dropping the previous parent, hence
        dropping this current node from the previous parent's children list.

        .. seealso::
            `orion.core.evc.tree.TreeNode.drop_parent`
        """
        if node is self.parent:
            return

        if node is not None and not isinstance(node, TreeNode):
            raise TypeError("Cannot set parent to %s" % str(node))

        if self.parent is not None:
            self.drop_parent()

        if node is not None:
            node.add_children(self)

    @property
    def children(self):
        """Get children of the node, empty list if no children"""
        return self._children

    def drop_children(self, *nodes):
        """Drop the children of the node, do nothing if no parent

        If no nodes are passed, the method will drop all the children of the current node.

        Note that the parent of the given node will be removed as well

        Raises
        ------
        ValueError
            If one of the given nodes is not a children of the current node.

        """
        if not nodes:
            nodes = self.children

        for child in list(nodes):
            del self._children[self._children.index(child)]
            # pylint: disable=protected-access
            child._parent = None

    def add_children(self, *nodes):
        """Add children to the current node

        Note that added children will have their parent set to the current node as well.

        .. seealso::
            `orion.core.evc.tree.TreeNode.drop_children`
        """
        for child in nodes:
            if child is not None and not isinstance(child, TreeNode):
                raise TypeError("Cannot add %s to children" % str(child))

            if child not in self._children:
                # TreeNode.set_parent uses add_children so using it here could cause an infinit
                # recursion. add_children() gets the dirty job done.
                child.drop_parent()
                # pylint: disable=protected-access
                child._parent = self
                self._children.append(child)

    @property
    def root(self):
        """Get the root of the tree

        Root node returns itself
        """
        if self.parent is None:
            return self

        return self.parent.root

    def map(self, function, node):
        r"""Apply a function recursively on the tree

        The function can be applied upwards on parents or downwards on children. The direction is
        defined by passing self.parent or self.children as the node argument.

        Parameters
        ----------
        function : callable
            Callable object to which will be passed the current node plus the parent node or the
            children nodes, depending on the direction of function application.
            If map on parents, callable(self, rval_parent_node)
            If map on children, callable(self, rval_children_nodes).
            Note that the callable object is expected to return an object which will be set as the
            current node's item (in the resulting tree), and the parent node or the children nodes
            depending on the direction of function application.
        node: None, `orion.core.evc.TreeNode` or list
            Can be either
            `None`: function is applied on current node only
            `self.parent`: function is applied recursively climbing up the tree until the root
            `self.children`: function is applied recursively going down the tree until the leafs

        Examples
        --------
        .. code-block:: python
            :linenos:

            # Tree structure
            # a
            # |   \  \   \
            # b    c  d   e
            # | \         |
            # f  g        h
            #

            root = TreeNode(1)
            b = TreeNode(1, root)
            TreeNode(1, root)
            TreeNode(1, root)
            e = TreeNode(1, root)

            f = TreeNode(1, b)
            TreeNode(1, b)

            h = TreeNode(1, e)

            def increment(node, children):
                for child in TreeNode(0, None, children):
                    child.item += 1

                return node.item + 1, children

            # Should return
            #
            # 2
            # |   \  \   \
            # 3    3  3   3
            # | \         |
            # 4  4        4

            rval = root.map(increment, root.children)
            assert [node.item for node in rval] == [2, 3, 4, 4, 3, 3, 3, 4]

            def increment_parent(node, parent):
                if parent is not None:
                    for parent in parent.root:
                        parent.item += 1

                return node.item + 1, parent

            # Should return
            #
            # 4
            # |
            # 3
            # |
            # 2

            rval = f.map(increment_parent, f.parent)
            assert [node.item for node in rval.root] == [4, 3, 2]

            rval = h.map(increment_parent, h.parent)
            assert [node.item for node in rval.root] == [4, 3, 2]

        """
        if node is None:
            rval, _ = function(self, None)
            return TreeNode(rval)
        elif node is self.parent:
            rval_parent_node = node.map(function, node.parent)
            rval, parent_node = function(self, rval_parent_node)
            return TreeNode(rval, parent_node)
        elif node is self.children:
            rval_children_nodes = [child.map(function, child.children) for child in self.children]
            rval, children_nodes = function(self, rval_children_nodes)
            return TreeNode(rval, parent=None, children=children_nodes)
        else:
            raise ValueError("Invalid nodes: %s" % str(node))

    def __iter__(self):
        """Iterate on the tree with pre-order traversal"""
        return PreOrderTraversal(self)

    def __repr__(self):
        """Represent the object as a string."""
        parent = self.parent
        if parent is not None:
            parent = parent.item

        children = [child.item for child in self.children]

        return ("%s(%s, parent=%s, children=%s)" %
                (self.__class__.__name__, str(self.item), str(parent), str(children)))


def flattened(trials_tree):
    """Get a list of the tree items in pre-order traversal"""
    return [node.item for node in trials_tree]
